"""Code to generate runoff data for the unit tests."""
import numpy as np
import pandas as pd

# set the seed
np.random.seed(42)

# making the test fraction of drainage area per state
# columns are 2-letter state abbreviations
# rows are fractions of areas
# sum across rows is 1
df_frac_da = pd.DataFrame({
    "07108900": np.random.random(4).tolist(),
    "07103980": np.random.random(4).tolist(),
    "01646500": np.random.random(4).tolist(),
}, index=["AL", "NY", "MN", "WI"]).T
# set one value to 0
df_frac_da.iloc[0, 0] = 0
# loop down rows and normalize
for i in range(3):
    df_frac_da.iloc[i, :] = \
        df_frac_da.iloc[i, :] / np.sum(df_frac_da.iloc[i, :])
# assert that the sum of each row is 1
assert np.allclose(np.sum(df_frac_da, axis=1), np.ones(3))

# make test fraction of state per drainage area
df_frac_state = pd.DataFrame({
    "AL": np.random.random(3).tolist(),
    "NY": np.random.random(3).tolist(),
    "MN": np.random.random(3).tolist(),
    "WI": np.random.random(3).tolist()
}, index=["07108900", "07103980", "01646500"])
# set one value to 0
df_frac_state.iloc[2, 1] = 0
# loop across columns and normalize
for i in range(4):
    df_frac_state.iloc[:, i] = \
        df_frac_state.iloc[:, i] / np.sum(df_frac_state.iloc[:, i])
# assert that the sum of each column is 1
assert np.allclose(np.sum(df_frac_state, axis=0), np.ones(4))

# make the weights table by doing element-wise multiplication
df_weights = df_frac_da * df_frac_state.values
# save the dataframe
df_weights.to_json("demo_weights.json")



"""Percentile calculation functions."""

import numpy as np
import pandas as pd
import warnings
from datetime import datetime
from hyswap.utils import filter_data_by_time
from hyswap.utils import calculate_metadata
from hyswap.utils import define_year_doy_columns
from hyswap.utils import set_data_type
from hyswap.utils import rolling_average


def calculate_fixed_percentile_thresholds(
        data,
        percentiles=np.array((0, 5, 10, 25, 75, 90, 95, 100)),
        method='weibull',
        ignore_na=True,
        **kwargs):
    """Calculate fixed percentile thresholds using historic data.

    Parameters
    ----------
    data : array_like
        1D array of data from which to calculate percentile thresholds.

    percentiles : array_like, optional
        Percentiles to calculate. Default is (0, 5, 10, 25, 75, 90, 95, 100).

    method : str, optional
        Method to use to calculate percentiles. Default is 'weibull'.

    ignore_na : bool, optional
        Ignore NA values in percentile calculations

    **kwargs : dict, optional
        Additional keyword arguments to pass to `numpy.percentile`.

    Returns
    -------
    percentiles : pandas.DataFrame
        Percentiles of the data in a DataFrame so the thresholds and
        percentile values are tied together.

    Examples
    --------
    Calculate default percentile thresholds from some synthetic data.

    .. doctest::

        >>> data = np.arange(101)
        >>> results = percentiles.calculate_fixed_percentile_thresholds(
        ...     data, method='linear')
        >>> results
        thresholds  0    5     10    25    75    90    95     100
        values      0.0  5.0  10.0  25.0  75.0  90.0  95.0  100.0

    Calculate a different set of thresholds from some synthetic data.

    .. doctest::

        >>> data = np.arange(101)
        >>> results = percentiles.calculate_fixed_percentile_thresholds(
        ...     data, percentiles=np.array((0, 10, 50, 90, 100)))
        >>> results
        thresholds  0    10    50    90     100
        values      0.0  9.2  50.0  90.8  100.0
    """
    if ignore_na:
        pct = np.nanpercentile(data, percentiles, method=method, **kwargs)
    else:
        pct = np.percentile(data, percentiles, method=method, **kwargs)
    df = pd.DataFrame(data={"values": pct}, index=percentiles).T
    df = df.rename_axis("thresholds", axis="columns")
    return df


def calculate_variable_percentile_thresholds_by_day(
        df,
        data_column_name,
        percentiles=[0, 5, 10, 25, 75, 90, 95, 100],
        method='weibull',
        date_column_name=None,
        data_type='daily',
        year_type='calendar',
        min_years=10,
        leading_values=0,
        trailing_values=0,
        ignore_na=True,
        **kwargs):
    """Calculate variable percentile thresholds of data by day of year.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing data to calculate daily percentile thresholds for.

    data_column_name : str
        Name of column containing data to analyze.

    percentiles : array_like, optional
        Percentile thresholds to calculate, default is
        [0, 5, 10, 25, 75, 90, 95, 100].

    method : str, optional
        Method to use to calculate percentiles. Default is 'weibull'.

    date_column_name : str, optional
        Name of column containing date information. If None, the index of
        `df` is used.

    data_type : str, optional
        The type of data. Must be one of 'daily', '7-day', '14-day', and
        '28-day'. Default is 'daily'. If '7-day', '14-day', or '28-day' is
        specified, the data will be averaged over the specified period. NaN
        values will be used for any days that do not have data. If present,
        NaN values will result in NaN values for the entire period.

    year_type : str, optional
        The type of year to use. Must be one of 'calendar', 'water', or
        'climate'. Default is 'calendar' which starts the year on January 1
        and ends on December 31. 'water' starts the year on October 1 and
        ends on September 30 of the following year which is the "water year".
        For example, October 1, 2010 to September 30, 2011 is "water year
        2011". 'climate' years begin on April 1 and end on March 31 of the
        following year, they are numbered by the ending year. For example,
        April 1, 2010 to March 31, 2011 is "climate year 2011".

    min_years : int, optional
        Minimum number of years of data required to calculate percentile
        thresholds for a given day of year. Default is 10.

    leading_values : int, optional
        For the temporal filtering, this is an argument setting the
        number of leading values to include in the output, inclusive.
        Default is 0, and parameter only applies to 'day' time_interval.

    trailing_values : int, optional
        For the temporal filtering, this is an argument setting the
        number of trailing values to include in the output, inclusive.
        Default is 0, and parameter only applies to 'day' time_interval.

    ignore_na : bool, optional
        Ignore NA values in percentile calculations

    **kwargs : dict, optional
        Additional keyword arguments to pass to `numpy.percentile`.

    Returns
    -------
    percentiles : pandas.DataFrame
        DataFrame containing threshold percentiles of data by day of year.
        Will return a DataFrame of NaNs for each percentile/day if
        provided an empty DataFrame or DataFrame with insufficient data

    Examples
    --------
    Calculate default thresholds by day of year from some real data in
    preparation for plotting.

    .. doctest::
        :skipif: True  # dataretrieval functions break CI pipeline

        >>> df, _ = dataretrieval.nwis.get_dv(
        ...     "03586500", parameterCd="00060",
        ...     start="1776-01-01", end="2022-12-31")
        >>> results = percentiles.calculate_variable_percentile_thresholds_by_day(  # noqa: E501
        ...     df, "00060_Mean")
        >>> len(results.index)  # 366 days in a leap year
        366
        >>> len(results.columns)  # 8 default percentiles
        8
    """
    # If the dataframe is empty, create a dummy dataframe to
    # run through function
    if df.empty:
        warnings.warn('No valid data provided, returning NA values for percentile thresholds')  # noqa: E501
        date_rng = pd.date_range(start='1900-01-01', end='1900-12-31')
        df = pd.DataFrame(index=date_rng)
        df[data_column_name] = np.nan

    # If data column name is not in dataframe
    if data_column_name not in df:
        warnings.warn('DataFrame missing data_column_name, returning NA values for percentile thresholds')  # noqa: E501
        date_rng = pd.date_range(start='1900-01-01', end='1900-12-31')
        df = pd.DataFrame(index=date_rng)
        df[data_column_name] = np.nan

    # define year and day of year columns and convert date column to datetime
    # if necessary
    df = define_year_doy_columns(df, date_column_name=date_column_name,
                                 year_type=year_type, clip_leap_day=True)
    # do rolling average for time as needed
    data_type = set_data_type(data_type)
    df = rolling_average(df, data_column_name, data_type)
    # based on date, get min and max day of year available
    min_day = np.nanmax((1, df.index.dayofyear.min()))
    max_day = np.nanmin((366, df.index.dayofyear.max() + 1))
    # make temporal index
    t_idx = np.arange(min_day, max_day)
    # initialize a DataFrame to hold percentiles by day of year
    percentiles_by_day = pd.DataFrame(index=t_idx, columns=percentiles)
    # loop through days of year available
    for doy in range(min_day, max_day):
        # get historical data for the day of year
        data = filter_data_by_time(df, doy, data_column_name,
                                   leading_values=leading_values,
                                   trailing_values=trailing_values,
                                   drop_na=ignore_na)
        if not data.empty:
            if not np.isnan(data).all():
                meta = calculate_metadata(data)
                # only calculate data if there are at least min_years of data
                # that are not nan
                if meta['n_years'] - meta['n_nans'] >= min_years:
                    # calculate percentiles for the day of year
                    # and add to DataFrame
                    _pct = calculate_fixed_percentile_thresholds(
                        data, percentiles=percentiles, method=method,
                        ignore_na=ignore_na, **kwargs)
                    percentiles_by_day.loc[t_idx == doy, :] = _pct.values.tolist()[0]  # noqa: E501
                else:
                    # if there are not at least 'min_years' of data,
                    # set percentiles to NaN
                    percentiles_by_day.loc[t_idx == doy, :] = np.nan
            else:
                # if all values are NA
                # set percentiles to NaN
                percentiles_by_day.loc[t_idx == doy, :] = np.nan
        else:
            # if the data subset for doy is empty
            # set percentiles to NaN
            percentiles_by_day.loc[t_idx == doy, :] = np.nan
    # replace index with multi-index of doy_index and month-day values
    # month_day values
    month_day = pd.to_datetime(
        percentiles_by_day.index, format='%j').strftime('%m-%d')
    # doy_index values
    doy_index = percentiles_by_day.index.values
    if year_type == 'water':
        doy_index = doy_index - 273
        doy_index[doy_index < 1] += 365
    elif year_type == 'climate':
        doy_index = doy_index - 90
        doy_index[doy_index < 1] += 365
    percentiles_by_day.index = pd.MultiIndex.from_arrays(
        [doy_index, month_day], names=['doy', 'month-day'])

    # sort percentiles by index
    percentiles_by_day.sort_index(inplace=True)

    # return percentiles by day of year
    return percentiles_by_day


def calculate_fixed_percentile_from_value(value, percentile_df):
    """Calculate percentile from a value and fixed percentile thresholds.

    This function enables faster calculation of the percentile associated with
    a given value if percentile values and corresponding fixed percentile
    thresholds are known from other data from the same station or site.
    This calculation is done using linear interpolation.

    Parameters
    ----------
    value : float, np.ndarray
        New value(s) to calculate percentile for. Can be a single value or an
        array of values.

    percentile_df : pd.DataFrame
        DataFrame where columns are the percentile thresholds values and the
        values are stored in a row called "values". Typically generated by the
        `calculate_fixed_percentile_thresholds` functions but could be
        provided manually or from data pulled from the NWIS stats service.

    Returns
    -------
    percentile : float, np.ndarray
        Percentile associated with the input value(s).

    Examples
    --------
    Calculate the percentile associated with a value from some synthetic data.

    .. doctest::

        >>> data = np.arange(1001)
        >>> pcts_df = percentiles.calculate_fixed_percentile_thresholds(
        ...     data, percentiles=[0, 5, 10, 25, 75, 90, 95, 100],
        ...     method='linear')
        >>> new_percentile = percentiles.calculate_fixed_percentile_from_value(
        ...     500, pcts_df)
        >>> new_percentile
        50.0

    Calculate the percentiles associated with multiple values for some data
    downloaded from NWIS.

    .. doctest::
        :skipif: True  # dataretrieval functions break CI pipeline

        >>> data, _ = dataretrieval.nwis.get_dv(
        ...     "04288000", parameterCd="00060",
        ...     start="1900-01-01", end="2021-12-31")
        >>> pcts_df = percentiles.calculate_fixed_percentile_thresholds(
        ...     data['00060_Mean'],
        ...     percentiles=[0, 5, 10, 25, 75, 90, 95, 100],
        ...     method='linear')
        >>> new_data, _ = dataretrieval.nwis.get_dv(
        ...     "04288000", parameterCd="00060",
        ...     start="2022-01-01", end="2022-01-07")
        >>> new_data['est_pct'] = percentiles.calculate_fixed_percentile_from_value(  # noqa: E501
        ...     new_data['00060_Mean'], pcts_df)
        >>> new_data['est_pct'].to_list()
        [58.41, 75.0, 48.45, 39.16, 45.58, 48.01, 42.04]

    """
    # define values
    thresholds = percentile_df.columns.tolist()
    percentile_values = percentile_df.values.tolist()[0]
    # do and return linear interpolation
    return np.interp(value, percentile_values, thresholds).round(2)


def calculate_variable_percentile_from_value(value, percentile_df, mo_day):
    """Calculate percentile from a value and variable percentile thresholds.

    This function enables faster calculation of the percentile associated with
    a given value for a single day of the year if percentile values and
    corresponding variable percentile thresholds are known from other data from
    the same station or site. This calculation is done using linear
    interpolation.

    Parameters
    ----------
    value : float, np.ndarray
        New value(s) to calculate percentile for. Can be a single value or an
        array of values.

    percentile_df : pd.DataFrame
        DataFrame containing threshold percentiles of data by day of year.
        Typically generated by the `calculate_variable_percentile_thresholds`
        function but could be provided manually or from data pulled from the
        NWIS stats service.

    mo_day : str
        string of month-day of year to lookup percentile thresholds for value

    Returns
    -------
    percentile : float, np.ndarray
        Percentile associated with the input value(s).

    Examples
    --------
    Calculate the percentile associated with a value using flow records
    downloaded from NWIS.

    .. doctest::
        :skipif: True  # dataretrieval functions break CI pipeline

        >>> data, _ = dataretrieval.nwis.get_dv(
        ...     "03586500", parameterCd="00060",
        ...     start="1776-01-01", end="2022-12-31")
        >>> pcts_df = percentiles.calculate_variable_percentile_thresholds_by_day(  # noqa: E501
        ...     data, '00060_Mean',
        ...     percentiles=[0, 5, 10, 25, 75, 90, 95, 100],
        ...     method='linear')
        >>> new_percentile = percentiles.calculate_variable_percentile_from_value(  # noqa: E501
        ...     500, pcts_df, '06-30')
        >>> new_percentile
        96.21
    """
    # retrieve percentile thresholds for the day of year of interest
    pct_values = percentile_df.loc[percentile_df.index.get_level_values('month-day') == mo_day]  # noqa: E501

    if not pct_values.empty:
        pct_values = pct_values.reset_index(drop=True)
        pct_values = pct_values.rename(index={0: "values"})
        # define values
        thresholds = pct_values.columns.tolist()
        percentile_values = pct_values.values.tolist()[0]
        # do and return linear interpolation
        est_pct = np.interp(value, percentile_values, thresholds).round(2)
    else:
        # return NaN if no threshold values are provided
        est_pct = np.nan

    return est_pct


def calculate_multiple_variable_percentiles_from_values(df, data_column_name,
                                                        percentile_df,
                                                        date_column_name=None):
    """Calculate variable percentiles for multiple values

    This function enables calculation of estimated percentiles for multiple
    values across multiple days of the year using existing variable percentile
    thresholds.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas dataframe containing new values to calculate percentiles for.

    data_column_name : str
        Name of column containing data to analyze.

    percentile_df : pd.DataFrame
        DataFrame containing threshold percentiles of data by day of year.
        Typically generated by the `calculate_variable_percentile_thresholds`
        functions but could be provided manually or from data pulled from the
        NWIS stats service.

    date_column_name : str, optional
        Name of column containing date information. If None, the index of
        `df` is used.

    Returns
    -------
    df : pd.DataFrame
        Pandas dataframe of values with estimated percentile column added

    Examples
    --------
    Calculate the percentiles associated with multiple values using flow
    records downloaded from NWIS.

    .. doctest::
        :skipif: True  # dataretrieval functions break CI pipeline

        >>> data, _ = dataretrieval.nwis.get_dv(
        ...     "04288000", parameterCd="00060",
        ...     start="1900-01-01", end="2021-12-31")
        >>> pcts_df = percentiles.calculate_variable_percentile_thresholds_by_day(  # noqa: E501
        ...     data, '00060_Mean',
        ...     percentiles=[0, 5, 10, 25, 75, 90, 95, 100],
        ...     method='linear')
        >>> new_data, _ = dataretrieval.nwis.get_dv(
        ...     "04288000", parameterCd="00060",
        ...     start="2022-01-01", end="2022-01-07")
        >>> new_percentiles = percentiles.calculate_multiple_variable_percentiles_from_values(  # noqa: E501
        ...     new_data, '00060_Mean', pcts_df)
        >>> new_percentiles['est_pct'].to_list()
        [59.59, 77.7, 47.5, 37.5, 50.0, 55.77, 48.71]
    """
    if date_column_name is None:
        date_column_name = 'datetime'

    df = df.reset_index()
    df['est_pct'] = df.apply(lambda row: calculate_variable_percentile_from_value(  # noqa: E501
        row[data_column_name], percentile_df,
        datetime.strftime(row[date_column_name], '%m-%d')), axis=1)
    df['est_pct'] = df['est_pct'].round(2)
    df = df.set_index(date_column_name)

    return df



"""Tests for the plotting functions."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from hyswap import exceedance
from hyswap import cumulative
from hyswap import rasterhydrograph
from hyswap import percentiles
from hyswap import plots
from hyswap import similarity


def test_plot_flow_duration_curve():
    """Test the plot_flow_duration_curve function."""
    values = [1, 1.25, 2.5]
    exceedance_probabilities = exceedance.calculate_exceedance_probability_from_values_multiple(  # noqa: E501
        values, [1, 2, 3, 4])
    ax = plots.plot_flow_duration_curve(values, exceedance_probabilities)
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Exceedance Probability\n' + \
        '(Percentage of time indicated value was equaled or exceeded)'
    assert ax.get_ylabel() == 'Discharge, ft3/s'
    assert ax.get_title() == 'Flow Duration Curve'
    assert len(ax.lines) == 1
    assert len(ax.collections) == 0
    assert len(ax.patches) == 0
    assert len(ax.texts) == 0
    assert len(ax.artists) == 0
    assert len(ax.images) == 0
    assert len(ax.containers) == 0
    assert len(ax.tables) == 0
    assert len(ax.spines) == 4
    # make another axes plot with different labels etc.
    ax = plots.plot_flow_duration_curve(values, exceedance_probabilities,
                                        title='Test Title',
                                        xlab='Test X Label',
                                        ylab='Test Y Label')
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Test X Label'
    assert ax.get_ylabel() == 'Test Y Label'
    assert ax.get_title() == 'Test Title'
    assert len(ax.lines) == 1
    assert len(ax.collections) == 0
    # make another plot with observations on it
    ax = plots.plot_flow_duration_curve(values, exceedance_probabilities,
                                        observations=[1, 2],
                                        observation_probabilities=[0.5, 0.25],
                                        title='Test Title',
                                        xlab='Test X Label',
                                        ylab='Test Y Label',
                                        scatter_kwargs={'c': 'k'})
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Test X Label'
    assert ax.get_ylabel() == 'Test Y Label'
    assert ax.get_title() == 'Test Title'
    assert len(ax.lines) == 1
    assert len(ax.collections) == 1
    # close plot
    plt.close()


def test_plot_raster_hydrograph():
    """Test the plot_raster_hydrograph function."""
    df = pd.DataFrame({'date': pd.date_range('1/1/2010', '12/31/2010'),
                       'data': np.random.rand(365)+100})
    df_formatted = rasterhydrograph.format_data(df, 'data', 'date')
    ax = plots.plot_raster_hydrograph(df_formatted)
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Month'
    assert ax.get_ylabel() == 'Year'
    assert ax.get_title() == 'Raster Hydrograph'
    # make one with custom labels
    ax = plots.plot_raster_hydrograph(df_formatted, title='Test Title',
                                      xlab='Test X Label',
                                      ylab='Test Y Label')
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Test X Label'
    assert ax.get_ylabel() == 'Test Y Label'
    assert ax.get_title() == 'Test Title'
    # close plot
    plt.close()


def test_plot_duration_hydrograph():
    """Test the plot_duration_hydrograph function."""
    df = pd.DataFrame({'date': pd.date_range('1/1/2010', '12/31/2012'),
                       'data': np.random.rand(1096)})
    df.index = df['date']
    pct = percentiles.calculate_variable_percentile_thresholds_by_day(
        df, 'data')
    df['doy'] = df.index.dayofyear
    ax = plots.plot_duration_hydrograph(pct, df, 'data', 'doy')
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Month'
    assert ax.get_ylabel() == 'Discharge, ft3/s'
    assert ax.get_title() == 'Duration Hydrograph'
    # make one with custom labels
    ax = plots.plot_duration_hydrograph(pct, df, 'data', 'doy',
                                        title='Test Title',
                                        xlab='Test X Label',
                                        ylab='Test Y Label',
                                        data_label='Test Data Label')
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Test X Label'
    assert ax.get_ylabel() == 'Test Y Label'
    assert ax.get_title() == 'Test Title'
    # close plot
    plt.close()


def test_plot_cumulative_hydrograph():
    """Test the plot_cumulative_hydrograph function."""
    df = pd.DataFrame({'date': pd.date_range('1/1/2010', '12/31/2010'),
                       'data': np.random.rand(365)})
    df_cumulative = cumulative.calculate_daily_cumulative_values(
        df, 'data', date_column_name='date')
    # apply plot function
    ax = plots.plot_cumulative_hydrograph(df_cumulative, 2010)
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Month'
    assert ax.get_ylabel() == 'Cumulative discharge, acre-feet'
    assert ax.get_title() == 'Cumulative Streamflow Hydrograph'
    assert len(ax.lines) == 1
    # make one with custom labels
    ax = plots.plot_cumulative_hydrograph(df_cumulative, 2010,
                                          title='Test Title',
                                          xlab='Test X Label',
                                          ylab='Test Y Label')
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Test X Label'
    assert ax.get_ylabel() == 'Test Y Label'
    assert ax.get_title() == 'Test Title'
    assert len(ax.lines) == 1
    # make one with min/max lines plotted
    ax = plots.plot_cumulative_hydrograph(df_cumulative, 2010,
                                          max_pct=True, min_pct=True)
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Month'
    assert ax.get_ylabel() == 'Cumulative discharge, acre-feet'
    assert ax.get_title() == 'Cumulative Streamflow Hydrograph'
    assert len(ax.lines) == 3
    assert len(ax.collections) == 1
    # make one with multiple years of data plotted
    _date = pd.date_range('1/1/2010', '12/31/2012')
    df = pd.DataFrame({'date': _date,
                       'data': np.random.rand(len(_date))})
    df_cumulative = cumulative.calculate_daily_cumulative_values(
        df, 'data', date_column_name='date')
    ax = plots.plot_cumulative_hydrograph(df_cumulative, [2010, 2011],
                                          max_pct=True, min_pct=True)
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Month'
    assert ax.get_ylabel() == 'Cumulative discharge, acre-feet'
    assert ax.get_title() == 'Cumulative Streamflow Hydrograph'
    assert len(ax.lines) == 4  # min/max, 2010 and 2011 lines
    assert len(ax.collections) == 1
    # close plot
    plt.close()


def test_plot_hydrograph():
    """Test the plot_hydrograph function."""
    # make some data
    df = pd.DataFrame({'date': pd.date_range('1/1/2010', '12/31/2010'),
                       'data': np.random.rand(365)})
    # apply plot function
    ax = plots.plot_hydrograph(df, 'data', 'date',
                               start_date='2/1/2010', end_date='10/30/2010')
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Date'
    assert ax.get_ylabel() == 'Discharge, ft3/s'
    assert ax.get_title() == 'Streamflow Hydrograph'
    assert len(ax.lines) == 1
    # close plot
    plt.close()


def test_plot_similarity_heatmat():
    """Testing the plot_similarity_heatmap function."""
    # make dummy dataframes
    df_1 = pd.DataFrame({
        'data': np.arange(10),
        'date': pd.date_range('2020-01-01', '2020-01-10')})
    df_1.set_index('date', inplace=True)
    df_2 = pd.DataFrame({
        'data': np.arange(10),
        'date': pd.date_range('2020-01-06', '2020-01-15')})
    df_2.set_index('date', inplace=True)
    # calculate similarity matrix (using correlation)
    df_corr, n_obs = similarity.calculate_correlations(
        [df_1, df_2], 'data')
    # plot
    ax = plots.plot_similarity_heatmap(df_corr, n_obs=n_obs,
                                       show_values=True,
                                       title='Test Title')
    # assertions
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Site'
    assert ax.get_ylabel() == 'Site'
    assert ax.get_title() == 'Test Title (n=5)'
    assert len(ax.lines) == 0
    # close plot
    plt.close()

    # make a set with negative values
    df_2 = pd.DataFrame({
        'data': np.random.random(10) * 10,
        'date': pd.date_range('2020-01-06', '2020-01-15')})
    df_2.set_index('date', inplace=True)
    # calculate similarity matrix (using correlation)
    df_corr, n_obs = similarity.calculate_correlations(
        [df_1, df_2], 'data')
    # plot
    ax = plots.plot_similarity_heatmap(df_corr, n_obs=n_obs,
                                       show_values=True,
                                       title='Test Title')
    # assertions
    assert isinstance(ax, plt.Axes)
    assert ax.get_xlabel() == 'Site'
    assert ax.get_ylabel() == 'Site'
    assert ax.get_title() == 'Test Title (n=5)'
    assert len(ax.lines) == 0
    # close plot
    plt.close()



"""Unit tests for the rasterhydrograph module."""
import pytest
import numpy as np
import pandas as pd
from hyswap import rasterhydrograph


class TestDateRange:
    def test_calculate_date_range_default(self):
        """Test the private function _calculate_date_range."""
        # define a data frame to use for testing
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates)),
             'index_year': dummy_dates.year.tolist(),
             'month': dummy_dates.month.tolist(),
             'day': dummy_dates.day.tolist(),
             'index_doy': dummy_dates.dayofyear.tolist()}
            )
        df.set_index('date', inplace=True)
        # test the function with default values
        date_range = rasterhydrograph._calculate_date_range(df, 'calendar',
                                                            None, None)
        assert date_range[0].year == 2018
        assert date_range[0].month == 1
        assert date_range[0].day == 1
        assert date_range[-1].year == 2022
        assert date_range[-1].month == 12
        assert date_range[-1].day == 31

    def test_calculate_date_range_specified(self):
        """Test the private function _calculate_date_range."""
        # define a data frame to use for testing
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates)),
             'index_year': dummy_dates.year.tolist(),
             'month': dummy_dates.month.tolist(),
             'day': dummy_dates.day.tolist(),
             'index_doy': dummy_dates.dayofyear.tolist()}
            )
        df.set_index('date', inplace=True)
        # test the function specifying start and end years
        date_range = rasterhydrograph._calculate_date_range(df, 'calendar',
                                                            2019, 2020)
        assert date_range[0].year == 2019
        assert date_range[0].month == 1
        assert date_range[0].day == 1
        assert date_range[-1].year == 2020
        assert date_range[-1].month == 12
        assert date_range[-1].day == 31


class TestCheckInputs:
    def test_check_inputs_01(self):
        """Test the private function _check_inputs."""
        # test the data frame input
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(None, None, None, None,
                                           None, None, None)

    def test_check_inputs_02(self):
        """Test the private function _check_inputs."""
        # test the data column name input
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), None, None, 'daily',
                                           'calendar', None, None)

    def test_check_inputs_03(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 1, None, 'daily',
                                           'calendar', None, None)

    def test_check_inputs_04(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 1.0, None, 'daily',
                                           'calendar', None, None)

    def test_check_inputs_05(self):
        """Test the private function _check_inputs."""
        # test the date column name input
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', {"a": 1},
                                           'daily', 'calendar', None, None)

    def test_check_inputs_06(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', 1, 'daily',
                                           'calendar', None, None)

    def test_check_inputs_07(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', 1.0,
                                           'daily', 'calendar', None, None)

    def test_check_inputs_08(self):
        """Test the private function _check_inputs."""
        # test the data type input
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 1,
                                           None, None, None)

    def test_check_inputs_09(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, 1.0,
                                           None, None, None)

    def test_check_inputs_10(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None, None,
                                           None, None, None)

    def test_check_inputs_11(self):
        """Test the private function _check_inputs."""
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'string', None, None, None)

    def test_check_inputs_12(self):
        """Test the private function _check_inputs."""
        # test the year type input
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', 1, None, None)

    def test_check_inputs_13(self):
        """Test the private function _check_inputs."""
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', 1.0, None, None)

    def test_check_inputs_14(self):
        """Test the private function _check_inputs."""
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', None, None, None)

    def test_check_inputs_15(self):
        """Test the private function _check_inputs."""
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', 'string', None, None)

    def test_check_inputs_16(self):
        """Test the private function _check_inputs."""
        # test the begin year input
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', 'calendar', 1.0, None)

    def test_check_inputs_17(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', 'calendar', 'string', None)

    def test_check_inputs_18(self):
        """Test the private function _check_inputs."""
        # make a df
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
            )
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(df, 'data', 'date', 'daily',
                                           'calendar', 2000, None)

    def test_check_inputs_19(self):
        """Test the private function _check_inputs."""
        # make a df
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
            )
        df_date = df.set_index('date')
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(df_date, 'data', None, 'daily',
                                           'calendar', 2000, None)

    def test_check_inputs_20(self):
        """Test the private function _check_inputs."""
        # test the end year input
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', 'calendar', None, 1.0)

    def test_check_inputs_21(self):
        """Test the private function _check_inputs."""
        with pytest.raises(TypeError):
            rasterhydrograph._check_inputs(pd.DataFrame(), 'data', None,
                                           'daily', 'calendar', None, 'string')

    def test_check_inputs_22(self):
        """Test the private function _check_inputs."""
        # make a df
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
            )
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(df, 'data', 'date', 'daily',
                                           'calendar', None, 3000)

    def test_check_inputs_23(self):
        """Test the private function _check_inputs."""
        # make a df
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
            )
        df_date = df.set_index('date')
        with pytest.raises(ValueError):
            rasterhydrograph._check_inputs(df_date, 'data', None, 'daily',
                                           'calendar', None, 3000)


class TestFormatData:
    def test_format_data(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        # test the function with a dataframe and a date column
        df_out = rasterhydrograph.format_data(df, 'value', 'date')
        assert len(df_out.index) == 5
        assert len(df_out.columns) == 365
        # assert day 1 of 2018 has no data, is NaN
        assert np.isnan(df_out.loc[2018][0])
        # assert day 350 of 2022 has no data, is NaN
        assert np.isnan(df_out.loc[2022][349])

    def test_format_data_date_index(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index
        df_out = rasterhydrograph.format_data(df_date, 'value')
        assert len(df_out.index) == 5
        assert len(df_out.columns) == 365
        # assert day 1 of 2018 has no data, is NaN
        assert np.isnan(df_out.loc[2018][0])
        # assert day 350 of 2022 has no data, is NaN
        assert np.isnan(df_out.loc[2022][349])

    def test_format_data_beginning_year(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and beginning year  # noqa
        df_out = rasterhydrograph.format_data(df_date, 'value', begin_year=2019)  # noqa
        assert len(df_out.index) == 4
        assert len(df_out.columns) == 365
        # assert day 1 of 2019 has data, is not NaN
        assert ~np.isnan(df_out.loc[2019][0])
        # assert day 350 of 2022 has no data, is NaN
        assert np.isnan(df_out.loc[2022][349])

    def test_format_data_end_year(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and ending year
        df_out = rasterhydrograph.format_data(df_date, 'value', end_year=2021)
        assert len(df_out.index) == 4
        assert len(df_out.columns) == 365
        # assert day 1 of 2018 has no data, is NaN
        assert np.isnan(df_out.loc[2018][0])
        # assert day 350 of 2021 has data, is not NaN
        assert ~np.isnan(df_out.loc[2021][349])

    def test_format_data_seven_day_averaging(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and a different
        # data averaging scheme
        df_7out = rasterhydrograph.format_data(df_date, 'value',
                                               data_type='7-day')
        assert len(df_7out.index) == 5
        assert len(df_7out.columns) == 365
        # assert day 1 of 2018 has no data, is NaN
        assert np.isnan(df_7out.loc[2018][0])
        # assert day 350 of 2022 has no data, is NaN
        assert np.isnan(df_7out.loc[2022][349])
        # check that there are non-NaN values in the data frame
        assert ~np.isnan(df_7out.values).all()

    def test_format_data_water_year(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and a water year
        df_out_water = rasterhydrograph.format_data(df_date, 'value',
                                                    year_type='water')
        df_out = rasterhydrograph.format_data(df_date, 'value', end_year=2021)
        assert len(df_out_water.index) == 5
        assert len(df_out_water.columns) == 365
        # check that there are non-NaN values in the data frame
        assert ~np.isnan(df_out_water.values).all()
        # check that day 1 of year 2019 is 10/1/2018
        assert df_out_water.loc[2019][0] == df_out.loc[2018][273]
        # check that the last day of year 2020 is 9/30/2020
        assert df_out_water.loc[2020][364] == df_out.loc[2020][272]

    def test_format_data_avg_water_year(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and a different
        # data averaging scheme and a water year
        df_7out_water = rasterhydrograph.format_data(df_date, 'value',
                                                     data_type='7-day',
                                                     year_type='water')
        df_7out = rasterhydrograph.format_data(df_date, 'value',
                                               data_type='7-day')
        assert len(df_7out_water.index) == 5
        assert len(df_7out_water.columns) == 365
        # check that there are non-NaN values in the data frame
        assert ~np.isnan(df_7out_water.values).all()
        # check that day 1 of year 2019 is 10/1/2018
        assert df_7out_water.loc[2019][0] == df_7out.loc[2018][273]
        # check that the last day of year 2020 is 9/30/2020
        assert df_7out_water.loc[2020][364] == df_7out.loc[2020][272]

    def test_format_data_climate_year(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and in a
        # climate year
        df_out_climate = rasterhydrograph.format_data(df_date, 'value',
                                                      year_type='climate')
        df_out = rasterhydrograph.format_data(df_date, 'value', end_year=2021)
        assert len(df_out_climate.index) == 4
        assert len(df_out_climate.columns) == 365
        # check that there are non-NaN values in the data frame
        assert ~np.isnan(df_out_climate.values).all()
        # check that day 1 of year 2020 is 4/1/2019
        assert df_out_climate.loc[2020][0] == df_out.loc[2019][90]
        # check that the last day of year 2021 is 3/31/2021
        assert df_out_climate.loc[2021][364] == df_out.loc[2021][89]

    def test_format_data_avg_climate_year(self):
        """Test the public function format_data."""
        # set up some data for the test
        dummy_dates = pd.date_range('2018-06-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and a different
        # data averaging scheme via kwargs and a climate year
        df_7out_climate = rasterhydrograph.format_data(df_date, 'value',
                                                       data_type='7-day',
                                                       year_type='climate',
                                                       center=True)
        assert len(df_7out_climate.index) == 4
        assert len(df_7out_climate.columns) == 365
        # check that there are non-NaN values in the data frame
        assert ~np.isnan(df_7out_climate.values).all()

    def test_format_data_climate_year_diff(self):
        """Test the public function format_data."""
        # this is a different averaging method so the values will be different
        # set up some data starting earlier in the year
        dummy_dates = pd.date_range('2018-02-01', '2022-01-31')
        df = pd.DataFrame(
            {'date': dummy_dates, 'value': np.random.rand(len(dummy_dates))}
        )
        df_date = df.set_index('date')
        # test the function with a dataframe and a date index and in a
        # climate year
        df_out_climate = rasterhydrograph.format_data(df_date, 'value',
                                                      year_type='climate')
        assert len(df_out_climate.index) == 5
        assert len(df_out_climate.columns) == 365
        # check that there are non-NaN values in the data frame
        assert ~np.isnan(df_out_climate.values).all()



"""Unit tests for the utils module."""
import pytest
import numpy as np
import pandas as pd
from hyswap import utils


class TestDataFilters:
    def test_filter_approved_data(self):
        """Test the filter_approved_data function."""
        data = {"a": ["A", "A", "P", "P"], "b": [1, 2, 3, 4]}
        data_df = pd.DataFrame(data)
        df = utils.filter_approved_data(data_df, filter_column="a")
        assert df["b"].tolist() == [1, 2]

    def test_filter_approved_data_error(self):
        """Test the filter_approved_data function."""
        data = {"a": ["A", "A", "P", "P"], "b": [1, 2, 3, 4]}
        data_df = pd.DataFrame(data)
        with pytest.raises(ValueError):
            utils.filter_approved_data(data_df, filter_column=None)


class TestRollingAverage:
    def test_rolling_average(self):
        """Test the rolling_average function."""
        # make a data frame with some dates and data
        df = pd.DataFrame(
            {"date": pd.date_range("2018-01-01", "2018-01-10"),
                "data": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})
        # set date as index
        df.set_index("date", inplace=True)
        # calculate the rolling average
        df_out = utils.rolling_average(df, "data", "2D")
        # check the output that is not na
        assert df_out["data"].dropna().tolist() == [1.5, 2.5, 3.5,
                                                    4.5, 5.5, 6.5,
                                                    7.5, 8.5, 9.5]
        # check that first element is NaN since day 1 does not have a 2-day
        # trailing window.
        assert np.isnan(df_out["data"][0])
        # check that the data column is in both dataframes
        assert "data" in df.columns
        assert "data" in df_out.columns
        # but check that the data are not equal
        assert df["data"].tolist() != df_out["data"].tolist()

    def test_rolling_average_kwarg_01(self):
        # make a data frame with some dates and data
        df = pd.DataFrame(
            {"date": pd.date_range("2018-01-01", "2018-01-10"),
                "data": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})
        # set date as index
        df.set_index("date", inplace=True)
        # try passing a kwarg to the rolling_average function
        df_out = utils.rolling_average(df, "data", "2D", center=True)
        assert df_out["data"].dropna().tolist() == [1.5, 2.5, 3.5,
                                                    4.5, 5.5, 6.5,
                                                    7.5, 8.5, 9.5]
        # check that last element is NaN since day 10 does not have a 2-day
        # leading window (since center=True, which changes the window
        # orientation).
        assert np.isnan(df_out["data"][9])

    def test_rolling_average_kwarg_02(self):
        # make a data frame with some dates and data
        df = pd.DataFrame(
            {"date": pd.date_range("2018-01-01", "2018-01-10"),
                "data": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})
        # set date as index
        df.set_index("date", inplace=True)
        # try passing a different kwarg
        df_out = utils.rolling_average(df, "data", "2D")
        assert np.isnan(df_out["data"].tolist()[0])
        assert df_out['data'].tolist()[1:] == [1.5, 2.5, 3.5, 4.5, 5.5, 6.5,
                                               7.5, 8.5, 9.5]


class TestFilterTime:
    def test_filter_data_by_time(self):
        """Test the filter_data_by_time function."""
        # make a dataframe
        df = pd.DataFrame({
            'data': [1, 2, 3, 4],
            'date': pd.date_range('2019-01-01', '2019-01-04')})
        # test the function
        data = utils.filter_data_by_time(df, 1, 'data',
                                         date_column_name='date')
        assert data.shape == (1,)
        assert data.values == [1]

    def test_filter_no_date(self):
        # make a dataframe
        df = pd.DataFrame({
            'data': [1, 2, 3, 4],
            'date': pd.date_range('2019-01-01', '2019-01-04')})
        # test the function with no date column and dates in the index
        df = df.set_index('date')
        data = utils.filter_data_by_time(df, 2, 'data')
        assert data.shape == (1,)
        assert data.values == [2]

    def test_filter_by_month(self):
        # test the filtering by month
        df = pd.DataFrame({
            'data': [1, 2, 3, 4],
            'date': pd.date_range('2019-01-01', '2019-05-01', freq='M')})
        data = utils.filter_data_by_time(df, 1, 'data',
                                         date_column_name='date',
                                         time_interval='month')
        assert data.shape == (1,)
        assert data.values == [1]

    def test_filter_by_year(self):
        # test the filtering by year
        df = pd.DataFrame({
            'data': [1, 2, 3, 4],
            'date': pd.date_range('2019-01-01', '2023-01-01', freq='Y')})
        data = utils.filter_data_by_time(df, 2019, 'data',
                                         date_column_name='date',
                                         time_interval='year')
        assert data.shape == (1,)
        assert data.values == [1]

    def test_filter_windowing(self):
        # test the windowing
        df = pd.DataFrame({
            'data': [1, 2, 3, 4],
            'date': pd.date_range('2019-01-01', '2019-01-04')})
        data = utils.filter_data_by_time(df, 2, 'data',
                                         date_column_name='date',
                                         leading_values=1)
        assert data.shape == (2,)
        assert np.all(data.values == [1, 2])

    def test_filter_tailing_window(self):
        # test tailing window
        df = pd.DataFrame({
            'data': [1, 2, 3, 4],
            'date': pd.date_range('2019-01-01', '2019-01-04')})
        data = utils.filter_data_by_time(df, 1, 'data',
                                         date_column_name='date',
                                         trailing_values=1)
        assert data.shape == (2,)
        assert np.all(data.values == [1, 2])

    def test_filter_error(self):
        df = pd.DataFrame({
            'data': [1, 2, 3, 4],
            'date': pd.date_range('2019-01-01', '2019-01-04')})
        # raise value error for invalid time interval
        with pytest.raises(ValueError):
            utils.filter_data_by_time(df, 1, 'data', date_column_name='date',
                                      time_interval='invalid')

    def test_filter_drop_na(self):
        # test ability to drop nans
        df = pd.DataFrame({
            'data': [1, 2, np.nan, 4],
            'date': pd.date_range('2019-01-01', '2019-01-04')})
        # test the function
        data = utils.filter_data_by_time(df, 1, 'data',
                                         date_column_name='date',
                                         time_interval='month',
                                         drop_na=False)
        data_no_nan = utils.filter_data_by_time(
            df, 1, 'data', date_column_name='date',
            time_interval='month', drop_na=True)
        # assertions
        assert np.isnan(data.values).sum() == 1
        assert np.isnan(data_no_nan.values).sum() == 0
        assert data.shape == (4,)
        assert data_no_nan.shape == (3,)


class TestMetadata:
    def test_calculate_metadata_01(self):
        """Test the calculate_metadata function."""
        # make pandas series of data with datetime index
        data = pd.Series(
            [1, 2, 3, 4],
            index=pd.date_range('2019-01-01', '2023-01-01', freq='Y'))
        # calculate the metadata
        metadata = utils.calculate_metadata(data)
        # check the output
        assert metadata['start_date'] == '2019-12-31'
        assert metadata['end_date'] == '2022-12-31'
        assert metadata['n_years'] == 4
        assert metadata['n_data'] == 4
        assert metadata['n_nans'] == 0
        assert metadata['n_zeros'] == 0
        assert metadata['n_missing_years'] == 0
        assert metadata['n_lows'] == 0

    def test_calculate_metadata_02(self):
        """Test the calculate_metadata function."""
        # make pandas series of data with datetime index
        data = pd.Series(
            [1, 2, 3, 4],
            index=pd.date_range('2019-01-01', '2023-01-01', freq='Y'))
        # remove the second value and recalculate the metadata
        data = data.drop(data.index[1])
        metadata = utils.calculate_metadata(data)
        # check the output
        assert metadata['start_date'] == '2019-12-31'
        assert metadata['end_date'] == '2022-12-31'
        assert metadata['n_years'] == 3
        assert metadata['n_data'] == 3
        assert metadata['n_nans'] == 0
        assert metadata['n_zeros'] == 0
        assert metadata['n_missing_years'] == 1
        assert metadata['n_lows'] == 0

    def test_calculate_metadata_03(self):
        """Test the calculate_metadata function."""
        # make pandas series of data with datetime index
        data = pd.Series(
            [1, 2, 3, 4],
            index=pd.date_range('2019-01-01', '2023-01-01', freq='Y'))
        # remove the second value
        data = data.drop(data.index[1])
        # make one value a nan and one a zero and recalculate the metadata
        data[0] = np.nan
        data[2] = 0
        metadata = utils.calculate_metadata(data)
        # check the output
        assert metadata['start_date'] == '2019-12-31'
        assert metadata['end_date'] == '2022-12-31'
        assert metadata['n_years'] == 3
        assert metadata['n_data'] == 2
        assert metadata['n_nans'] == 1
        assert metadata['n_zeros'] == 1
        assert metadata['n_missing_years'] == 1
        assert metadata['n_lows'] == 1


class TestDefiningColumns:
    def test_define_year_doy_columns(self):
        """Test the define_year_doy_columns function."""
        # make dummy df
        df = pd.DataFrame({
            'data': np.arange(1, 366),
            'date': pd.date_range('2019-01-01', '2019-12-31')})
        # apply function
        df = utils.define_year_doy_columns(df, date_column_name='date')
        assert df['index_year'].tolist() == [2019] * 365
        assert df['index_doy'].tolist() == list(range(1, 366))
        assert df.index.dayofyear.tolist() == list(range(1, 366))

    def test_define_year_doy_columns_water_year(self):
        # make dummy df
        df = pd.DataFrame({
            'data': np.arange(1, 366),
            'date': pd.date_range('2019-01-01', '2019-12-31')})
        # test with a different year type
        df = utils.define_year_doy_columns(df, date_column_name='date',
                                           year_type='water')
        assert df.index.year.unique().tolist() == [2019]
        assert df['index_year'].unique().tolist() == [2019, 2020]
        assert df['index_doy'].tolist() == list(range(93, 366)) + \
            list(range(1, 93))

    def test_define_year_doy_columns_climate_year(self):
        # make dummy df
        df = pd.DataFrame({
            'data': np.arange(1, 366),
            'date': pd.date_range('2019-01-01', '2019-12-31')})
        # test with climate year
        df = utils.define_year_doy_columns(df, date_column_name='date',
                                           year_type='climate')
        assert df.index.year.unique().tolist() == [2019]
        assert df['index_year'].unique().tolist() == [2019, 2020]
        assert df['index_doy'].tolist() == list(range(276, 366)) + \
            list(range(1, 276))

    def test_define_year_doy_columns_leap_year(self):
        # test with leap year
        df = pd.DataFrame({
            'data': np.arange(1, 367),
            'date': pd.date_range('2020-01-01', '2020-12-31')})
        df = utils.define_year_doy_columns(df, date_column_name='date',
                                           clip_leap_day=False)
        assert df.index.year.unique().tolist() == [2020]
        assert df['index_year'].unique().tolist() == [2020]
        assert df['index_doy'].tolist() == list(range(1, 367))
        assert len(df['index_doy']) == 366

    def test_define_year_doy_columns_leap_clip(self):
        # test with leap year and clip leap day
        df = pd.DataFrame({
            'data': np.arange(1, 367),
            'date': pd.date_range('2020-01-01', '2020-12-31')})
        df = utils.define_year_doy_columns(df, date_column_name='date',
                                           clip_leap_day=True)
        assert df.index.year.unique().tolist() == [2020]
        assert df['index_year'].unique().tolist() == [2020]
        assert df['index_doy'].tolist() == list(range(1, 366))
        assert len(df['index_doy']) == 365

    def test_define_year_doy_columns_water_year_clip(self):
        # water year leap with clip
        df = pd.DataFrame({
            'data': np.arange(1, 367),
            'date': pd.date_range('2020-01-01', '2020-12-31')})
        df = utils.define_year_doy_columns(df, date_column_name='date',
                                           year_type='water',
                                           clip_leap_day=True)
        assert df.index.year.unique().tolist() == [2020]
        assert df['index_year'].unique().tolist() == [2020, 2021]
        assert df['index_doy'].tolist() == list(range(93, 366)) + \
            list(range(1, 93))
        assert len(df['index_doy']) == 365

    def test_define_year_doy_columns_climate_year_clip(self):
        # climate year leap with clip
        df = pd.DataFrame({
            'data': np.arange(1, 367),
            'date': pd.date_range('2020-01-01', '2020-12-31')})
        df = utils.define_year_doy_columns(df, date_column_name='date',
                                           year_type='climate',
                                           clip_leap_day=True)
        assert df.index.year.unique().tolist() == [2020]
        assert df['index_year'].unique().tolist() == [2020, 2021]
        assert df['index_doy'].tolist() == list(range(276, 366)) + \
            list(range(1, 276))
        assert len(df['index_doy']) == 365


class TestMungingNWISStats:
    def test_munge_nwis_stats(self):
        """Test the munge_nwis_stats function."""
        # make a dataframe
        df = pd.DataFrame({
            'month_nu': [1, 2, 3, 4],
            'day_nu': [1, 2, 3, 4],
            'end_yr': [2019, 2019, 2019, 2019],
            'min_va': [1, 2, 3, 4],
            'max_va': [5, 6, 7, 8],
            'mean_va': [9, 10, 11, 12],
            'p05_va': [13, 14, 15, 16],
            'p10_va': [17, 18, 19, 20],
            'p20_va': [21, 22, 23, 24],
            'p25_va': [25, 26, 27, 28],
            'p50_va': [29, 30, 31, 32],
            'p75_va': [33, 34, 35, 36],
            'p80_va': [37, 38, 39, 40],
            'p90_va': [41, 42, 43, 44],
            'p95_va': [45, 46, 47, 48],
        })
        # apply the function
        df_slim = utils.munge_nwis_stats(df)
        # check the output
        assert df_slim.shape == (4, 11)
        assert len(df.columns) > len(df_slim.columns)
        assert 0 in df_slim.columns
        assert df_slim.columns.tolist() == [0, 5, 10, 20, 25, 50, 75, 80, 90,
                                            95, 100]

    def test_munge_additional_params(self):
        # make a dataframe
        df = pd.DataFrame({
            'month_nu': [1, 2, 3, 4],
            'day_nu': [1, 2, 3, 4],
            'end_yr': [2019, 2019, 2019, 2019],
            'min_va': [1, 2, 3, 4],
            'max_va': [5, 6, 7, 8],
            'mean_va': [9, 10, 11, 12],
            'p05_va': [13, 14, 15, 16],
            'p10_va': [17, 18, 19, 20],
            'p25_va': [21, 22, 23, 24],
            'p50_va': [25, 26, 27, 28],
            'p75_va': [29, 30, 31, 32],
            'p90_va': [33, 34, 35, 36],
            'p95_va': [37, 38, 39, 40],
        })
        # function w/ additional parameters
        df_slim = utils.munge_nwis_stats(df, ['min_va', 'max_va'], [0, 100])
        assert df_slim.shape == (4, 2)
        assert df_slim.columns.tolist() == [0, 100]
        assert len(df.columns) > len(df_slim.columns)

    def test_munge_raise_error(self):
        # make a dataframe
        df = pd.DataFrame({
            'month_nu': [1, 2, 3, 4],
            'day_nu': [1, 2, 3, 4],
            'end_yr': [2019, 2019, 2019, 2019],
            'min_va': [1, 2, 3, 4],
            'max_va': [5, 6, 7, 8],
            'mean_va': [9, 10, 11, 12],
            'p05_va': [13, 14, 15, 16],
            'p10_va': [17, 18, 19, 20],
            'p25_va': [21, 22, 23, 24],
            'p50_va': [25, 26, 27, 28],
            'p75_va': [29, 30, 31, 32],
            'p90_va': [33, 34, 35, 36],
            'p95_va': [37, 38, 39, 40],
        })
        # raise error if column lists are of different lengths
        with pytest.raises(ValueError):
            utils.munge_nwis_stats(df, ['min_va', 'max_va'], [0, 100, 200])


def test_filter_to_common_time():
    """Test the filter_to_common_time function."""
    # make dummy dataframes
    df_1 = pd.DataFrame({
        'data': np.arange(10),
        'date': pd.date_range('2020-01-01', '2020-01-10')})
    df_1.set_index('date', inplace=True)
    df_2 = pd.DataFrame({
        'data': np.arange(10),
        'date': pd.date_range('2020-01-06', '2020-01-15')})
    df_2.set_index('date', inplace=True)
    # apply the function
    df_list, n_obs = utils.filter_to_common_time([df_1, df_2])
    # check the output
    assert len(df_list) == 2
    assert n_obs == 5


def test_set_data_type():
    """Test the function set_data_type."""
    assert utils.set_data_type('daily') == '1D'
    assert utils.set_data_type('7-day') == '7D'
    assert utils.set_data_type('14-day') == '14D'
    assert utils.set_data_type('28-day') == '28D'


class TestSummaryStatistics:
    def test_calculate_summary_statistics(self):
        """Test the calculate_summary_statistics function."""
        # make test dataframe
        df = pd.DataFrame({
            'datetime': pd.date_range('2000-01-01', '2000-01-10'),
            '00060_Mean': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            'site_no': np.ones(10) * 12345678
        })
        # set datetime as index
        df.set_index('datetime', inplace=True)
        # use function
        df_stats = utils.calculate_summary_statistics(df, '00060_Mean')
        # check output
        assert df_stats.shape == (8, 1)
        assert df_stats.columns[0] == 'Summary Statistics'
        assert df_stats.iloc[0, 0] == '12345678'
        assert df_stats.iloc[1, 0] == '2000-01-01'
        assert df_stats.iloc[2, 0] == '2000-01-10'
        assert df_stats.iloc[3, 0] == 10
        assert df_stats.iloc[4, 0] == 1
        assert df_stats.iloc[5, 0] == 5.5
        assert df_stats.iloc[6, 0] == 5.5
        assert df_stats.iloc[7, 0] == 10


class TestSetDataType:
    def test_set_data_type_daily(self):
        """Test the function set_data_type."""
        assert utils.set_data_type('daily') == '1D'

    def test_set_data_type_week(self):
        """Test the function set_data_type."""
        assert utils.set_data_type('7-day') == '7D'

    def test_set_data_type_two_week(self):
        """Test the function set_data_type."""
        assert utils.set_data_type('14-day') == '14D'

    def test_set_data_type_twenty_eight_day(self):
        """Test the function set_data_type."""
        assert utils.set_data_type('28-day') == '28D'


class TestCategorizationSchema:

    def test_retrieve_schema_exists(self):
        """Test the function retrieve_schema."""
        schema = utils.retrieve_schema('NWD')
        assert isinstance(schema, dict)
        assert 'ranges' in schema.keys()
        assert 'labels' in schema.keys()

    def test_retrieve_schema_missing(self):
        """Test the function retrieve_schema."""
        with pytest.raises(ValueError):
            utils.retrieve_schema('abcdef')

    def test_retrieve_schema_casesensitive(self):
        """Test the function retrieve_schema."""
        schema = utils.retrieve_schema('Nwd')
        assert isinstance(schema, dict)


class TestFlowCategorization:
    # define a basic categorization schema
    schema = {'ranges': [0, 25, 75, 100],
              'labels': ['Low', 'Med', 'High'],
              'low_label': 'Lowest',
              'high_label': 'Highest'}

    def test_flow_categorization(self):
        """Test the function flow_categorization."""
        # define a simple dataframe of flow percentiles
        df = pd.DataFrame({
            'datetime': pd.date_range('2000-01-01', '2000-01-06'),
            'est_pct': [10, 50, 74.999, 75, 75.0001, 90]})
        df.set_index('datetime', inplace=True)
        df = utils.categorize_flows(df, 'est_pct', custom_schema=self.schema)
        assert 'flow_cat' in df.columns
        assert df.iloc[0, -1] == 'Low'
        assert df.iloc[1, -1] == 'Med'
        assert df.iloc[2, -1] == 'Med'
        assert df.iloc[3, -1] == 'High'
        assert df.iloc[4, -1] == 'High'
        assert df.iloc[5, -1] == 'High'

    def test_flow_categorization_lowest_highest(self):
        """Test the function flow_categorization."""
        # define a simple dataframe of flow percentiles
        df = pd.DataFrame({
            'datetime': pd.date_range('2000-01-01', '2000-01-02'),
            'est_pct': [0, 100]})
        df.set_index('datetime', inplace=True)
        df = utils.categorize_flows(df, 'est_pct', custom_schema=self.schema)
        assert 'flow_cat' in df.columns
        assert df.iloc[0, -1] == 'Lowest'
        assert df.iloc[1, -1] == 'Highest'

    def test_flow_categorization_nan(self):
        """Test the function flow_categorization."""
        # define a simple dataframe of flow percentiles that are not valid
        df = pd.DataFrame({
            'datetime': pd.date_range('2000-01-01', '2000-01-03'),
            'est_pct': [-1, np.nan, 101]})
        df.set_index('datetime', inplace=True)
        df = utils.categorize_flows(df, 'est_pct', custom_schema=self.schema)
        assert df['flow_cat'].isnull().all()



"""Tests for the percentiles functions."""
import numpy as np
import pandas as pd
import pytest
from hyswap import percentiles


class TestCalculateFixedPercentileThresholds:
    # data value for all tests in this class
    data = np.arange(101)

    def test_calculate_fixed_percentile_thresholds_defaults(self):
        """Test the calculate_fixed_percentile_thresholds function defaults."""
        # test the function
        percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
            self.data, method='linear', ignore_na=False)
        assert percentiles_.shape == (1, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.values.tolist()[0] == [
            0.0, 5.0, 10.0, 25.0, 75.0, 90.0, 95.0, 100.0]

    def test_custom_percentiles(self):
        # set some percentile values as opposed to the defaults
        percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
            self.data, percentiles=np.array((0, 10, 50, 90, 100)))
        assert percentiles_.shape == (1, 5)
        assert percentiles_.columns.tolist() == [0, 10, 50, 90, 100]
        assert percentiles_.values.tolist()[0] == [
            0.0, 9.200000000000001, 50.0, 90.80000000000001, 100.0]

    def test_kwargs_to_percentile(self):
        # pass kwarg through to np.percentile
        percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
            self.data, method='lower')
        assert percentiles_.shape == (1, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.values.tolist()[0] == [
            0, 5, 10, 25, 75, 90, 95, 100]

    def test_with_nans(self):
        # test with some nan values
        percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
            [np.nan] + list(self.data) + [np.nan, np.nan],
            method='linear', ignore_na=True)
        assert percentiles_.shape == (1, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.values.tolist()[0] == [
            0.0, 5.0, 10.0, 25.0, 75.0, 90.0, 95.0, 100.0]

    def test_empty_df(self):
        # test with an empty dataframe
        percentiles_ = percentiles.calculate_fixed_percentile_thresholds(
            pd.DataFrame([]),
            method='linear', ignore_na=True)
        assert percentiles_.shape == (1, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.isna().all().all()


class TestCalculateVariablePercentileThresholdsByDay:
    # data for all tests in this class
    small_df = pd.DataFrame({
        'data': np.arange(101),
        'date': pd.date_range('2019-01-01', '2019-04-11')})

    bigger_df = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2000-01-01', '2020-12-31'))),
        'date': pd.date_range('2000-01-01', '2020-12-31')})

    df_dummy = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2000-01-01', '2001-12-31'))),
        'date': pd.date_range('2000-01-01', '2001-12-31')})

    df_dummy_alt = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2001-01-01', '2002-12-31'))),
        'date': pd.date_range('2001-01-01', '2002-12-31')})

    # This dataframe has 2 years of data but is missing values
    # in the month of May.
    include_months = [1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12]
    df_dummy_missing_days = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2001-01-01', '2002-12-31'))),
        'date': pd.date_range('2001-01-01', '2002-12-31')})
    df_dummy_missing_days = df_dummy_missing_days[df_dummy_missing_days['date'].dt.month.isin(include_months)] # noqa

    # This dataframe has 10 years of data, but 5 days have nans
    # in year 1.
    df_nans = pd.DataFrame({
        'data': np.random.random(
            len(pd.date_range('2001-01-01', '2010-12-31'))),
        'date': pd.date_range('2001-01-01', '2010-12-31')})
    df_nans.data.iloc[0:5] = np.nan

    def test_calculate_variable_percentile_thresholds_by_day(self):
        """Test with date column."""
        # test the function
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.small_df, 'data', date_column_name='date')
        assert percentiles_.shape == (101, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
        assert percentiles_.index.get_level_values(0).tolist()[0] == 1

    def test_calculate_variable_percentile_thresholds_by_day_dateindex(self):
        """Test with dates in the index."""
        # test the function with no date column and dates in the index
        self.small_df = self.small_df.set_index('date')
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.small_df, 'data')
        assert percentiles_.shape == (101, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
        assert percentiles_.index.get_level_values(0).tolist()[0] == 1

    def test_calculate_variable_percentile_thresholds_by_day_percentiles(self):
        # test the function with a different set of percentiles
        self.small_df = self.small_df.set_index('date')
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.small_df, 'data', percentiles=np.array((0, 10, 50, 90, 100)))
        assert percentiles_.shape == (101, 5)
        assert percentiles_.columns.tolist() == [0, 10, 50, 90, 100]
        assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
        assert percentiles_.index.get_level_values(0).tolist()[0] == 1
        # all percentiles should be NaN because demo dataset is 1 year only
        assert percentiles_.isna().all().all()

    def test_calculate_variable_percentile_thresholds_by_day_year_type(self):
        """Test with a different year type."""
        # test the function with a different year type
        self.small_df = self.small_df.set_index('date')
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.small_df, 'data', year_type='water')
        assert percentiles_.shape == (101, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.index.get_level_values(1).tolist()[0] == '01-01'
        assert percentiles_.index.get_level_values(0).tolist()[0] == 93
        assert percentiles_.index.get_level_values(1).tolist()[-1] == '04-11'
        assert percentiles_.index.get_level_values(0).tolist()[-1] == 193
        # all percentiles should be NaN because demo dataset is 1 year only
        assert percentiles_.isna().all().all()

    def test_calculate_variable_percentile_thresholds_by_day_year_type_percentiles(self): # noqa
        # test the function with a different year type and a different set of
        # percentiles
        self.small_df = self.small_df.set_index('date')
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.small_df, 'data', year_type='climate',
            percentiles=np.array((0, 10, 50, 90, 100)))
        assert percentiles_.shape == (101, 5)
        assert percentiles_.columns.tolist() == [0, 10, 50, 90, 100]
        assert percentiles_.index.get_level_values(1).tolist()[0] == '04-01'
        assert percentiles_.index.get_level_values(0).tolist()[0] == 1
        # all percentiles should be NaN because demo dataset is 1 year only
        assert percentiles_.isna().all().all()

    def test_non_nans(self):
        """Test that the percentiles are not NaN."""
        # test the function and check that the percentiles are not NaN
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.bigger_df, 'data', date_column_name='date')
        assert not percentiles_.isna().all().all()

    def test_longer_dummy_set(self):
        """Test with a longer dummy set."""
        # test a longer dummy set that exceeds 1 year
        # test the function
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.df_dummy, 'data', date_column_name='date', year_type='water')
        assert percentiles_.shape == (365, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.index.get_level_values(1).tolist()[0] == '10-01'
        assert percentiles_.index.get_level_values(0).tolist()[0] == 1
        assert percentiles_.index.get_level_values(1).tolist()[-1] == '09-30'
        assert percentiles_.index.get_level_values(0).tolist()[-1] == 365

    def test_longer_dummy_alt(self):
        """Test with a different longer dummy set."""
        # test a longer dummy set that exceeds 1 year
        # test the function
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.df_dummy_alt, 'data', date_column_name='date',
            year_type='water')
        assert percentiles_.shape == (365, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_.index.get_level_values(1).tolist()[0] == '10-01'
        assert percentiles_.index.get_level_values(0).tolist()[0] == 1
        assert percentiles_.index.get_level_values(1).tolist()[-1] == '09-30'
        assert percentiles_.index.get_level_values(0).tolist()[-1] == 365

    def test_bigger_rolling_avg(self):
        """Test rolling average."""
        # function w/ daily percentiles
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.bigger_df, 'data', date_column_name='date')
        percentiles_7day = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.bigger_df, 'data', date_column_name='date',
            data_type='7-day')
        assert percentiles_.shape == (365, 8)
        assert percentiles_7day.shape == (365, 8)
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]
        assert percentiles_7day.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100] # noqa
        # check that the percentiles are not NaN
        assert not percentiles_.isna().all().all()
        assert not percentiles_7day.isna().all().all()
        # check that the percentiles are not the same
        assert not percentiles_.equals(percentiles_7day)

    def test_empty_df(self):
        """Test that function returns empty percentile df."""
        # test the function with an empty DataFrame and check
        # that the percentiles are all NaN
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            pd.DataFrame([]), 'data')
        assert percentiles_.isna().all().all()
        assert percentiles_.columns.tolist() == [0, 5, 10, 25, 75, 90, 95, 100]

    def test_missing_doys(self):
        """Test that function returns empty percentile df."""
        # test the function can handle when days of the year
        # are missing from the dataframe. These days should
        # not be included or show up as NaN
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.df_dummy_missing_days, 'data', date_column_name='date',
            min_years=1)
        assert not percentiles_.isna().all().all()
        assert percentiles_.iloc[126].isna().all()
        assert percentiles_.shape == (365, 8)
        # assert percentiles_[percentiles_.index.get_level_values('doy') == 5].isna().all() # noqa

    def test_nan_years_variable_percentiles_calculations(self):
        """Test that function returns empty percentiles for insufficient years due to nans."""  # noqa
        percentiles_ = percentiles.calculate_variable_percentile_thresholds_by_day( # noqa
            self.df_nans, 'data', date_column_name='date')
        assert percentiles_.iloc[0:4].isna().all().all()


class TestCalculateFixedPercentilesFromValue:
    # define some test values
    data = np.arange(101)
    percentiles_ = np.arange(0, 105, 5)
    pct_df = pd.DataFrame(data={"values": percentiles_}, index=percentiles_).T
    pct_df = pct_df.rename_axis("thresholds", axis="columns")
    low_val = -5
    high_val = 105
    mid_val = 51
    multiple_values = [-5, 7, 51, 98, 105]

    def test_calculate_fixed_percentiles_from_value_low(self):
        """Test with a low value."""
        # test the function
        pct_out = percentiles.calculate_fixed_percentile_from_value(
            self.low_val, self.pct_df)
        assert pct_out == 0.0

    def test_calculate_fixed_percentiles_from_value_mid(self):
        """Test with a mid value."""
        # test the function
        pct_out = percentiles.calculate_fixed_percentile_from_value(
            self.mid_val, self.pct_df)
        assert pct_out == 51.0

    def test_calculate_fixed_percentiles_from_value_high(self):
        """Test with a high value."""
        # test the function
        pct_out = percentiles.calculate_fixed_percentile_from_value(
            self.high_val, self.pct_df)
        assert pct_out == 100.0

    def test_calculate_fixed_percentiles_from_value_multiple(self):
        """Test with multiple values."""
        pct_out = percentiles.calculate_fixed_percentile_from_value(
            self.multiple_values, self.pct_df)
        assert pct_out == pytest.approx([0.0, 7.0, 51.0, 98.0, 100.0])

    def test_calculate_fixed_percentiles_from_value_invalid(self):
        """Test with invalid inputs."""
        with pytest.raises(AttributeError):
            percentiles.calculate_fixed_percentile_from_value(
                self.low_val, [1, 2, 3])



"""Runoff functions for hyswap."""
import pandas as pd


def convert_cfs_to_runoff(cfs, drainage_area, frequency="annual"):
    """Convert cfs to runoff values for some drainage area.

    Parameters
    ----------
    cfs : float
        Flow in cubic feet per second.

    drainage_area : float
        Drainage area in km2.

    frequency : str, optional
        Frequency of runoff values to return. Options are 'annual',
        'monthly', and 'daily'. Default is 'annual'.

    Returns
    -------
    float
        Runoff in mm/<frequency>.

    Examples
    --------
    Convert 14 cfs to mm/yr for a 250 km2 drainage area.

    .. doctest::

        >>> mmyr = runoff.convert_cfs_to_runoff(14, 250)
        >>> np.round(mmyr)
        50.0
    """
    # convert frequency string to value
    if frequency == "annual":
        freq = 365.25
    elif frequency == "monthly":
        freq = 365.25 / 12
    elif frequency == "daily":
        freq = 1
    else:
        raise ValueError("Invalid frequency: {}".format(frequency))
    # convert cfs to cubic feet per frequency
    cpf = cfs * 60 * 60 * 24 * freq
    # convert cubic feet per freq to cubic meters per freq
    cpf = cpf * (0.3048 ** 3)
    # convert drainage area km2 to m2
    drainage_area = drainage_area * 1000 * 1000
    # convert cubic meters per year to mm per freq
    mmf = cpf / drainage_area * 1000
    return mmf


def streamflow_to_runoff(df, data_col, drainage_area, frequency="annual"):
    """Convert streamflow to runoff for a given drainage area.

    For a given gage/dataframe, convert streamflow to runoff using the
    drainage area and the convert_cfs_to_runoff function.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing streamflow data.

    data_col : str
        Column name containing streamflow data, assumed to be in cfs.

    drainage_area : float
        Drainage area in km2.

    frequency : str, optional
        Frequency of runoff values to return. Options are 'annual',
        'monthly', and 'daily'. Default is 'annual'.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing runoff data in a column named 'runoff'.

    Examples
    --------
    Convert streamflow to runoff for a given drainage area.

    .. doctest::

        >>> df = pd.DataFrame({'streamflow': [14, 15, 16]})
        >>> runoff_df = runoff.streamflow_to_runoff(df, 'streamflow', 250)
        >>> print(runoff_df['runoff'].round(1))
        0    50.0
        1    53.6
        2    57.2
        Name: runoff, dtype: float64
    """
    df['runoff'] = df[data_col].apply(
        lambda x: convert_cfs_to_runoff(x, drainage_area, frequency=frequency)
    )
    return df


def calculate_geometric_runoff(geom_id, df_list, weights_matrix,
                               start_date=None, end_date=None,
                               data_col='runoff'):
    """Function to calculate the runoff for a specified geometry.

    Parameters
    ----------
    geom_id : str
        Geometry ID for the geometry of interest.

    df_list : list
        List of dataframes containing runoff data for each site in the
        geometry.

    weights_matrix : pandas.DataFrame
        DataFrame containing the weights for all sites and all geometries.
        Columns are geometry IDs, index is site IDs.

    start_date : str, optional
        Start date for the runoff calculation. If not specified, the earliest
        date in the df_list will be used. Format is 'YYYY-MM-DD'.

    end_date : str, optional
        End date for the runoff calculation. If not specified, the latest
        date in the df_list will be used. Format is 'YYYY-MM-DD'.

    data_col : str, optional
        Column name containing runoff data in the dataframes in df_list.
        Default is 'runoff', as it is assumed these dataframes are created
        using the :obj:`streamflow_to_runoff` function.

    Returns
    -------
    pandas.Series
        Series containing the area-weighted runoff values for the geometry.
    """
    # get date range
    date_range = _get_date_range(df_list, start_date, end_date)

    # get site list from weights matrix index
    site_list = weights_matrix.index.tolist()

    # create empty dataframe to store results, index is site_list,
    # columns are date_range
    runoff_df = pd.DataFrame(index=site_list, columns=date_range)

    # loop through the df_list to populate rows of the runoff_df
    for df in df_list:
        # get site id (assumed to be string from NWIS)
        site_id = df['site_no'][0]
        # convert site_id to int for indexing weights matrix
        site_id = int(site_id)
        # get weight for site
        weight = weights_matrix[geom_id].loc[site_id]
        # get runoff for site
        runoff = df[data_col]
        # multiply weights by runoff
        weighted_runoff = weight * runoff
        # add weighted runoff to runoff_df
        # pandas seems to use dates to automatically align data :)
        runoff_df.loc[site_id] = weighted_runoff

    # combine the new runoff_df with the existing weights matrix to calculate
    # the area-weighted runoff values for the geometry
    runoff_sum = runoff_df.sum(axis=0, skipna=True)
    weights_sum = weights_matrix[geom_id].sum(skipna=True)
    weighted_runoff = runoff_sum / weights_sum

    return weighted_runoff


def _get_date_range(df_list, start_date, end_date):
    """Get date range for runoff calculation.

    This is an internal function used by the :obj:`calculate_geometric_runoff`
    function to get the date range for the runoff calculation. If no start or
    end date is specified, the earliest/latest date in the df_list will be
    used.

    Parameters
    ----------
    df_list : list
        List of dataframes containing runoff data for each site in the
        geometry.

    start_date : str, optional
        Start date for the runoff calculation. If not specified, the earliest
        date in the df_list will be used. Format is 'YYYY-MM-DD'.

    end_date : str, optional
        End date for the runoff calculation. If not specified, the latest
        date in the df_list will be used. Format is 'YYYY-MM-DD'.

    Returns
    -------
    pandas.DatetimeIndex
        DatetimeIndex containing the date range for the runoff calculation.
    """
    if start_date is None:
        # if no start date iterate through df_list to find earliest date
        start_date = df_list[0].index[0]
        for df in df_list:
            if df.index[0] < start_date:
                start_date = df.index[0]
    else:
        # make the string tz-aware
        start_date = pd.to_datetime(start_date).tz_localize("UTC")
    if end_date is None:
        # if no end date iterate through df_list to find latest date
        end_date = df_list[0].index[-1]
        for df in df_list:
            if df.index[-1] > end_date:
                end_date = df.index[-1]
    else:
        # make the string tz-aware
        end_date = pd.to_datetime(end_date).tz_localize("UTC")
    # create range of dates from start to end
    date_range = pd.date_range(start_date, end_date)

    return date_range


def identify_sites_from_weights(geom_id,
                                weights_df,
                                geom_id_col,
                                site_col,
                                wght_in_basin_col='pct_in_basin',
                                wght_in_geom_col='pct_in_huc'):

    """Identify sites for a specified geometry.

    Function to identify sites with non-zero weights for a given
    spatial geometry. This function is a helper function that can
    be used to reduce the number of NWIS queries that are performed
    to construct the list of dataframes for a given geometry.

    Parameters
    ----------
    geom_id : str
        Geometry ids to filter to (e.g. geom_id = '03030006').
        Ids range from 8 to 10 digits and sometime involve a leading 0.

    weights_df : pandas.DataFrame
        Tabular dataFrame containing columns the site numbers,
        geometry ids, and two columns wghts in huc and the drainage area basin.

    geom_id_col : str
        Column in weights_df with geometry ids.

    site_col: str
        Column in weights_df with drainage area site numbers.
        Please make sure ids have the correct number of digits and have
          not lost leading 0s when read in.
        If the site numbers are the weights_df index col, site_col = 'index'.

    wght_in_basin_col: float, optional
        Name of column with values representing the proportion (0 to 1)
        of the spatial geometry occurring in the corresponding drainage area.
        Default name: pct_in_basin)

    wght_in_geom_col: float, optional
        Name of column with values representing the proportion (0 to 1)
        of the drainage area occurring in the corresponding spatial geometry.
        Default name: pct_in_huc


    Returns
    -------
    list
        List of site IDs with non-zero weights for the geometry

    Examples
    --------
    Find all drainage area site numbers that intersect with geometry id 0101002

    .. doctest::

        >>> data = [['01014000', '01010002', 0.01, 0.6],
        ... ['01014001', '01010002', 0.2, 0.8],
        ... ['01014002', '01010003', 0.9, 0.05]]
        >>> df = pd.DataFrame(data,
        ... columns = ['site_no', 'geom_id', 'wght_basin', 'wght_huc'])
        >>> sites_lst = runoff.identify_sites_from_weights(weights_df=df,
        ... geom_id='01010002', geom_id_col='geom_id', site_col='site_no',
        ... wght_in_basin_col='wght_basin', wght_in_geom_col='wght_huc')
        >>> print(sites_lst)
        ['01014000', '01014001']

    """

    # Filter df to designated geometry (e.g. huc8)
    filtered_df = weights_df[weights_df[geom_id_col] == geom_id]

    # Check that all sites ids have a character count of at least 8
    assert all(filtered_df[site_col].str.len() >= 8), (
        'site numbers char length must be > or = to 8 char')

    # Check whether sites is the df index or not
    if site_col == 'index':
        site_col = filtered_df.index
    else:
        site_col = filtered_df[site_col]

    # Retrieve all non-0 sites within the designated geometry (e.g. huc8)
    site_list = site_col[(filtered_df[wght_in_basin_col] != 0) | (
        filtered_df[wght_in_geom_col] != 0)].to_list()

    return site_list


def calculate_multiple_geometric_runoff(
        geom_id_list, df_list, weights_matrix,
        start_date=None, end_date=None, data_col='runoff'):
    """Calculate runoff for multiple geometries at once.

    Parameters
    ----------
    geom_id : list
        List of geometry ID strings for the geometries of interest.
        These should be columns in the weights matrix.

    df_list : list
        List of dataframes containing runoff data for each site in the
        geometry.

    weights_matrix : pandas.DataFrame
        DataFrame containing the weights for all sites and all geometries.
        Columns are geometry IDs, index is site IDs.

    start_date : str, optional
        Start date for the runoff calculation. If not specified, the earliest
        date in the df_list will be used. Format is 'YYYY-MM-DD'.

    end_date : str, optional
        End date for the runoff calculation. If not specified, the latest
        date in the df_list will be used. Format is 'YYYY-MM-DD'.

    data_col : str, optional
        Column name containing runoff data in the dataframes in df_list.
        Default is 'runoff', as it is assumed these dataframes are created
        using the :obj:`streamflow_to_runoff` function.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the area-weighted runoff values for each
        geometry. Columns are geometry IDs, index is date range.
    """
    # create empty dataframe to store results
    results_df = pd.DataFrame()
    # loop through geom_id_list to calculate runoff for each geometry
    for geom_id in geom_id_list:
        # calculate runoff for geometry
        runoff = calculate_geometric_runoff(
            geom_id, df_list, weights_matrix,
            start_date=start_date, end_date=end_date,
            data_col=data_col)
        # add runoff to results_df
        results_df[geom_id] = runoff.to_frame()
    return results_df



"""Utility functions for hyswap."""
import pandas as pd


def filter_approved_data(data, filter_column=None):
    """Filter a dataframe to only return approved "A" (or "A, e") data.

    Parameters
    ----------
    data : pandas.DataFrame
        The data to filter.
    filter_column : string
        The column upon which to filter. If None, an error will be raised.

    Returns
    -------
    pandas.DataFrame
        The filtered data.

    Examples
    --------
    Filter synthetic data to only return approved data. First make some
    synthetic data.

    .. doctest::

        >>> df = pd.DataFrame({
        ...     'data': [1, 2, 3, 4],
        ...     'approved': ['A', 'A', 'P', 'P']})
        >>> df.shape
        (4, 2)

    Then filter the data to only return approved data.

    .. doctest::

        >>> df = utils.filter_approved_data(df, filter_column='approved')
        >>> df.shape
        (2, 2)
    """
    if filter_column is None:
        raise ValueError("filter_column must be specified.")
    return data.loc[((data[filter_column] == "A") |
                     (data[filter_column] == "A, e") |
                     (data[filter_column] == "A, R"))]


def rolling_average(df, data_column_name, data_type,
                    auto_min_periods=True, custom_min_periods=None,
                    **kwargs):
    """Calculate a rolling average for a dataframe.

    Default behavior right-aligns the window used for the rolling average
    and uses the data_type argument ('1D', '7D', '14D', '28D') to set the
    `min_periods` argument in `pandas.DataFrame.rolling`. The function
    returns NaN values if any of the values in the window are NaN or if the
    `mind_periods` argument is not satisifed. Properties of the windowing
    can be changed by passing additional keyword arguments which are fed
    to `pandas.DataFrame.rolling`.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to calculate the rolling average for.
    data_column_name : string
        The name of the column to calculate the rolling average for.
    data_type : string
        The formatted frequency string to be used with
        pandas.DataFrame.rolling to calculate the average over the correct
        temporal period.
    auto_min_periods : bool
        Defaults to True. When True, the `min_periods` argument in
        `pandas.DataFrame.rolling` is set using the `data_type` argument.
        For example, if the `data_type` = '7D', the `min_periods`
        argument is 7. When False, the `min_periods` argument is set
        using the `custom_min_periods` input.
    custom_min_periods : int, optional
        Defaults to None. Only used if `auto_min_periods` is False.
        If `auto_min_periods` is False and an integer is provided,
        that integer will be used to define the `min_periods` argument
        in `pandas.DataFrame.rolling`.
    **kwargs
        Additional keyword arguments to be passed to
        `pandas.DataFrame.rolling`.

    Returns
    -------
    pandas.DataFrame
        The output dataframe with the rolling average values.
    """
    if auto_min_periods is True:
        min_periods = pd.to_timedelta(data_type).days
    else:
        min_periods = custom_min_periods
    df_out = df.copy(deep=True)
    df_out[data_column_name] = df_out[data_column_name].rolling(
        data_type,
        min_periods, **kwargs).mean()
    return df_out


def filter_data_by_time(df, value, data_column_name, date_column_name=None,
                        time_interval='day',
                        leading_values=0, trailing_values=0,
                        drop_na=False):
    """Filter data by some time interval.

    DataFrame containing data to filter. Expects datetime information to be
    available in the index or in a column named `date_column_name`. The
    returned `pandas.Series` object will have the datetimes for the specified
    time (day, month, year) as the index, and the corresponding data from the
    `data_column_name` column as the values.

    Parameters
    ----------
    value : int
        Time value to use for filtering; value can be a day of year (1-366),
        month (1-12), or year (4 digit year).

    data_column_name : str
        Name of column containing data to filter.

    date_column_name : str, optional
        Name of column containing date information. If None, the index of
        `df` is used.

    time_interval : str, optional
        Time interval to filter by. Must be one of 'day', 'month', or 'year'.
        Default is 'day'.

    leading_values : int, optional
        Number of leading values to include in the output, inclusive.
        Default is 0, and parameter only applies to 'day' time_interval.

    trailing_values : int, optional
        Number of trailing values to include in the output, inclusive.
        Default is 0, and parameter only applies to 'day' time_interval.

    drop_na : bool, optional
        Drop NA values within filtered data

    Returns
    -------
    data : pandas.Series
        Data from the specified day of year.

    Examples
    --------
    Filter some synthetic data by day of year. First make some synthetic data.

    .. doctest::

        >>> df = pd.DataFrame({
        ...     'data': [1, 2, 3, 4],
        ...     'date': pd.date_range('2019-01-01', '2019-01-04')})
        >>> df.shape
        (4, 2)

    Then filter the data to get data from day 1.

    .. doctest::

        >>> data = utils.filter_data_by_time(
        ...     df, 1, 'data', date_column_name='date')
        >>> data.shape
        (1,)

    Acquire and filter some real daily data to get all Jan. 1 data.

    .. doctest::
        :skipif: True  # dataretrieval functions break CI pipeline

        >>> df, _ = dataretrieval.nwis.get_dv(
        ...     "03586500", parameterCd="00060",
        ...     start="2000-01-01", end="2003-01-05")
        >>> data = utils.filter_data_by_time(df, 1, '00060_Mean')
        >>> data.shape
        (4,)
    """
    # make date column the index if it is not already
    if date_column_name is not None:
        df = df.set_index(date_column_name)
    # check that time_interval is valid
    if time_interval not in ['day', 'month', 'year']:
        raise ValueError(
            'time_interval must be one of "day", "month", or "year".')
    if time_interval == 'day':
        if (leading_values == 0) and (trailing_values == 0):
            # grab data from the specified day of year
            dff = df.loc[df.index.dayofyear == value, data_column_name]
        else:
            # grab data from the specified day of year and include leading
            # and trailing values.
            # note that at the beginning and end of the year, this section
            # wraps backward and forward, respectively, to ensure it is
            # calculating percentiles from a full window.
            if value < (1 + leading_values):
                dff = df.loc[
                    (df.index.dayofyear >= value - leading_values) &
                    (df.index.dayofyear <= value + trailing_values) |
                    (df.index.dayofyear >= 366 - (leading_values - value)),
                    data_column_name]
            elif value > (366 - trailing_values):
                dff = df.loc[
                    (df.index.dayofyear >= value - leading_values) &
                    (df.index.dayofyear <= value + trailing_values) |
                    (df.index.dayofyear <= trailing_values - (366 - value)),
                    data_column_name]
            else:
                dff = df.loc[
                    (df.index.dayofyear >= value - leading_values) &
                    (df.index.dayofyear <= value + trailing_values),
                    data_column_name]
    elif time_interval == 'month':
        # grab data from the specified month
        dff = df.loc[df.index.month == value, data_column_name]
    elif time_interval == 'year':
        # grab data from the specified year
        dff = df.loc[df.index.year == value, data_column_name]
    if drop_na:
        dff = dff.dropna()
    # return data as a pandas Series where the index is the date
    return dff


def calculate_metadata(data):
    """Calculate metadata for a series of data.

    Parameters
    ----------
    data : pandas.Series
        The data to calculate the metadata for. Expected to have a datetime
        index.

    Returns
    -------
    dict
        The calculated metadata which includes the number of years of data,
        the number of data points, any gaps in the data, and the start and end
        dates of the data, the number of 0 values, the number of NA values,
        as well as the number of low (typically low flow <= 0.01) values.
    """
    # initialize the metadata dictionary
    meta = {}
    # calculate the number of unique years of data
    meta["n_years"] = len(data.index.year.unique())
    # calculate the number of data points that are not nan
    meta["n_data"] = len(data.loc[~data.isna()])
    # calculate the number of gaps in the data - missing years
    expected_years = data.index.year.max() - data.index.year.min() + 1
    meta["n_missing_years"] = expected_years - meta["n_years"]
    # calculate the start and end dates of the data
    meta["start_date"] = data.index.min().strftime("%Y-%m-%d")
    meta["end_date"] = data.index.max().strftime("%Y-%m-%d")
    # calculate the number of 0 values
    meta["n_zeros"] = len(data.loc[data == 0])
    # calculate the number of nan values
    meta["n_nans"] = len(data.loc[data.isna()])
    # calculate the number of low values (below 0.01)
    meta["n_lows"] = len(data.loc[data <= 0.01])

    return meta


def define_year_doy_columns(df_in, date_column_name=None, year_type='calendar',
                            clip_leap_day=False):
    """Function to add year, day of year, and month-day columns to a DataFrame.

    Parameters
    ----------
    df_in : pandas.DataFrame
        DataFrame containing data to filter. Expects datetime information to be
        available in the index or in a column named `date_column_name`.

    date_column_name : str, optional
        Name of column containing date information. If None, the index of
        `df` is used.

    year_type : str, optional
        The type of year to use. Must be one of 'calendar', 'water', or
        'climate'. Default is 'calendar' which starts the year on January 1
        and ends on December 31. 'water' starts the year on October 1 and
        ends on September 30 of the following year which is the "water year".
        For example, October 1, 2010 to September 30, 2011 is "water year
        2011". 'climate' years begin on April 1 and end on March 31 of the
        following year, they are numbered by the ending year. For example,
        April 1, 2010 to March 31, 2011 is "climate year 2011".

    clip_leap_day : bool, optional
        If True, February 29 is removed from the DataFrame. Default is False.

    Returns
    -------
    df : pandas.DataFrame
        DataFrame with year, day of year, and month-day columns added. Also
        makes the date_column_name the index of the DataFrame.
    """
    # deep copy of the df before manipulating it
    df = df_in.copy(deep=True)
    # set the df index
    if date_column_name is not None:
        df = df.set_index(date_column_name)
    # check that year_type is valid
    if year_type not in ['calendar', 'water', 'climate']:
        raise ValueError(
            'year_type must be one of "calendar", "water", or "climate".')
    # add year and day of year columns
    if year_type == 'calendar':
        df['index_year'] = df.index.year
        df['index_doy'] = df.index.dayofyear
    elif year_type == 'water':
        # set water years
        df['index_year'] = df.index.year.where(df.index.month < 10,
                                               df.index.year + 1)
        # get calendar day of year
        df['index_doy'] = df.index.dayofyear
        # adjust Oct 1 to be day 1 of water year for leap and non-leap years
        df.loc[df.index.is_leap_year & (df.index.month >= 10),
               'index_doy'] -= 274
        df.loc[~df.index.is_leap_year & (df.index.month >= 10),
               'index_doy'] -= 273
        # adjust Jan 1 accordingly for leap and non-leap years
        df.loc[df.index.month < 10, 'index_doy'] += 92
    elif year_type == 'climate':
        # set climate years
        df['index_year'] = df.index.year.where(df.index.month < 4,
                                               df.index.year + 1)
        # get calendar day of year
        df['index_doy'] = df.index.dayofyear
        # adjust Apr 1 to be day 1 of climate year for leap and non-leap years
        df.loc[df.index.is_leap_year & (df.index.month >= 4),
               'index_doy'] -= 91
        df.loc[~df.index.is_leap_year & (df.index.month >= 4),
               'index_doy'] -= 90
        # adjust Jan 1 to be day 276 of climate year for all years
        df.loc[df.index.month < 4, 'index_doy'] += 275
    # add month and day columns
    df['index_month_day'] = df.index.strftime('%m-%d')
    # clip leap year and adjustment
    if clip_leap_day:
        df = leap_year_adjustment(df, year_type=year_type)
    # sort the df by year and day of year
    df = df.sort_values(['index_year', 'index_doy'])
    return df


def leap_year_adjustment(df, year_type='calendar'):
    """Function to adjust leap year days in a DataFrame.

    Adjust for a leap year by removing February 29 from the DataFrame and
    adjusting the day of year values for the remaining days of the year.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing data to adjust. Expects datetime information to be
        available in the index and a column named 'doy' containing day of year.

    year_type : str, optional
        The type of year to use. Must be one of 'calendar', 'water', or
        'climate'. Default is 'calendar' which starts the year on January 1
        and ends on December 31. 'water' starts the year on October 1 and
        ends on September 30 of the following year which is the "water year".
        For example, October 1, 2010 to September 30, 2011 is "water year
        2011". 'climate' years begin on April 1 and end on March 31 of the
        following year, they are numbered by the ending year. For example,
        April 1, 2010 to March 31, 2011 is "climate year 2011".

    Returns
    -------
    df : pandas.DataFrame
        DataFrame with leap year days removed and day of year values adjusted.
    """
    df = df.loc[~((df.index.month == 2) & (df.index.day == 29))]
    if year_type == 'calendar':
        df.loc[df.index.is_leap_year & (df.index.month > 2),
               'index_doy'] -= 1
    elif year_type == 'water':
        df.loc[df.index.is_leap_year &
               (df.index.month > 2) &
               (df.index.month < 10), 'index_doy'] -= 1
    elif year_type == 'climate':
        df.loc[df.index.is_leap_year &
               (df.index.month > 2) &
               (df.index.month < 4), 'index_doy'] -= 1
    return df


def munge_nwis_stats(df, source_pct_col=None, target_pct_col=None):
    """Function to munge and reformat NWIS statistics data.

    This is a utility function that exists to help munge NWIS percentile data
    served via the NWIS statistics web service. This function is intended to
    be used on Python dataretrieval dataframe returns for the nwis.get_stats()
    function for "daily" data, a single site, and a single parameter code.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing NWIS statistics data retrieved from the statistics
        web service. Assumed to come in as a dataframe retrieved with a
        package like dataretrieval or similar.
    source_pct_col : list, optional
        List of column names to use as the source percentiles. If None, the
        values are assumed to correspond to the 0, 5, 10, 20, 25, 50, 75, 80,
        90, 95, and 100 percentiles in the NWIS statistics service return.
    target_pct_col : list, optional
        List of column names to use as the target percentiles. If None, then
        integer values are used as the column names corresponding to the
        default source values.

    Returns
    -------
    df_slim : pandas.DataFrame
        DataFrame containing munged and reformatted NWIS statistics data.
        Reformatting is for use with the hyswap package plotting function for
        duration hydrographs with statistical information in background of
        the plot.

    Examples
    --------
    Get some NWIS statistics data.

    .. doctest::

        >>> df, md = dataretrieval.nwis.get_stats(
        ...     "03586500", parameterCd="00060", statReportType="daily")

    Then apply the function to munge the data.

    .. doctest::

        >>> df = utils.munge_nwis_stats(df)
        >>> df.shape
        (366, 11)
    """
    # set defaults
    if source_pct_col is None:
        source_pct_col = ['min_va', 'p05_va', 'p10_va', 'p20_va', 'p25_va',
                          'p50_va', 'p75_va', 'p80_va', 'p90_va', 'p95_va',
                          'max_va']
    if target_pct_col is None:
        target_pct_col = [0, 5, 10, 20, 25, 50, 75, 80, 90, 95, 100]
    # check lengths of lists for column names
    if len(source_pct_col) != len(target_pct_col):
        raise ValueError('source_pct_col and target_pct_col must be the same '
                         'length')
    # rename date columns
    df.rename(columns={'month_nu': 'month', 'day_nu': 'day', 'end_yr': 'year'},
              inplace=True)
    # make end year 2020 for leap year
    df['year'] = 2020
    # construct date column
    df['date'] = pd.to_datetime(df[['day', 'month', 'year']])
    # set month-day as index
    df['month_day'] = df['date'].dt.strftime('%m-%d')
    df = df.set_index('month_day')
    # slim down to just the columns used for the plot
    df_slim = df[source_pct_col]
    # rename columns
    df_slim.columns = target_pct_col
    # return the dataframe
    return df_slim


def calculate_summary_statistics(df, data_col="00060_Mean"):
    """
    Calculate summary statistics for a site.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing daily values for the site. Expected to be from
        `dataretrieval.nwis.get_dv()`, or similar.

    data_col : str, optional
        Name of the column in the dv_df DataFrame that contains the data of
        interest. Default is "00060_Mean" which is the mean daily discharge
        column.

    Returns
    -------
    summary_df : pandas.DataFrame
        DataFrame containing summary statistics for the site.

    Examples
    --------
    Get some NWIS data and apply the function to get the summary statistics.

    .. doctest::

        >>> df, _ = dataretrieval.nwis.get_dv(
        ...     "03586500", parameterCd="00060",
        ...     startDT="2010-01-01", endDT="2010-12-31")
        >>> summary_df = utils.calculate_summary_statistics(df)
        >>> summary_df.shape
        (8, 1)
        >>> print(summary_df)
                    Summary Statistics
        Site number           03586500
        Begin date          2010-01-01
        End date            2010-12-31
        Count                      365
        Minimum                   2.48
        Mean                    207.43
        Median                    82.5
        Maximum                 3710.0
    """
    # make dictionary
    summary_dict = {}
    # populate it
    # site number
    summary_dict['Site number'] = str(int(df['site_no'][0])).zfill(8)
    # dates
    summary_dict['Begin date'] = df.index.min().strftime('%Y-%m-%d')
    summary_dict['End date'] = df.index.max().strftime('%Y-%m-%d')
    # count
    summary_dict['Count'] = df[data_col].count()
    # minimum
    summary_dict['Minimum'] = df[data_col].min()
    # mean
    summary_dict['Mean'] = df[data_col].mean().round(2)
    # median
    summary_dict['Median'] = df[data_col].median()
    # maximum
    summary_dict['Maximum'] = df[data_col].max()

    # make dataframe
    summary_df = pd.DataFrame(summary_dict, index=[0])

    # transpose and set column name
    summary_df = summary_df.T
    summary_df.columns = ['Summary Statistics']

    # return dataframe
    return summary_df


def filter_to_common_time(df_list):
    """Filter a list of dataframes to common times based on index.

    This function takes a list of dataframes and filters them to only include
    the common times based on the index of the dataframes. This is necessary
    before comparing the timeseries and calculating statistics between two or
    more timeseries.

    Parameters
    ----------
    df_list : list
        List of pandas.DataFrame objects to filter to common times.
        DataFrames assumed to have date-time information in the index.
        Expect input to be the output from a function like
        dataretrieval.nwis.get_dv() or similar.

    Returns
    -------
    df_list : list
        List of pandas.DataFrame objects filtered to common times.
    n_obs : int
        Number of observations in the common time period.

    Examples
    --------
    Get some NWIS data.

    .. doctest::

            >>> df1, md1 = dataretrieval.nwis.get_dv(
            ...     "03586500", parameterCd="00060",
            ...     start="2018-12-15", end="2019-01-07")
            >>> df2, md2 = dataretrieval.nwis.get_dv(
            ...     "01646500", parameterCd="00060",
            ...     start="2019-01-01", end="2019-01-14")
            >>> type(df1)
            <class 'pandas.core.frame.DataFrame'>
            >>> type(df2)
            <class 'pandas.core.frame.DataFrame'>

    Filter the dataframes to common times.

    .. doctest::

            >>> df_list, n_obs = utils.filter_to_common_time([df1, df2])
            >>> df_list[0].shape
            (7, 3)
            >>> df_list[1].shape
            (7, 3)
    """
    # get the common index
    common_index = df_list[0].index
    for df in df_list:
        common_index = common_index.intersection(df.index)
    # filter the dataframes to the common index
    for i, df in enumerate(df_list):
        df_list[i] = df.loc[common_index]
    # get the number of observations
    n_obs = len(common_index)
    # return the list of dataframes
    return df_list, n_obs


def set_data_type(data_type):
    """Function to set the data type for rolling averages.

    Parameters
    ----------
    data_type : str
        The type of data. Must be one of 'daily', '7-day', '14-day', and
        '28-day'. If '7-day', '14-day', or '28-day' is
        specified, the data will be averaged over the specified period. NaN
        values will be used for any days that do not have data. If present,
        NaN values will result in NaN values for the entire period.

    Returns
    -------
    data_type : str
        The formatted frequency string to be used with
        pandas.DataFrame.rolling to calculate the average over the correct
        temporal period.
    """
    if data_type == 'daily':
        data_type = '1D'
    elif data_type == '7-day':
        data_type = '7D'
    elif data_type == '14-day':
        data_type = '14D'
    elif data_type == '28-day':
        data_type = '28D'

    return data_type


def categorize_flows(df, data_col, schema_name='NWD', custom_schema=None):
    """Function to categorize streamflows based on percentile ranges

    This function assigns a category to each streamflow observation for a
    single site by comparing the estimated percentile to a schema of percentile
    ranges and associated category labels

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing values for the site.

    data_col : str
        Name of the column in the DataFrame that contains the data of
        interest, in this case estimated streamflow percentile.

    schema_name : str, optional
        Name of the categorization schema that should be used to categorize
        streamflow. Default is "NWD" schema.

    custom_schema : dict, optional
        Python dictionary describing custom schema to use for categorizing
        streamflow based on percentiles. Required in dict is 'ranges', an array
        of percentile cut points and 'labels', a list of category labels that
        matches the number of bins represented by ranges. Optionally can
        include 'low_label' and 'high_label' which are category labels
        associated with the lowest and highest values in 'ranges',
        respectively. Additional optional keys include 'colors', 'low_color',
        and 'high_color' which specify a color palette that can be accessed in
        user created plots and maps. Default is None.

    Returns
    -------
    df : pandas.DataFrame
        DataFrame with flow_cat column added.

    Examples
    --------
    Categorize streamflow based on calculated percentiles for streamflow
    records downloaded from NWIS.

    .. doctest::
        :skipif: True  # dataretrieval functions break CI pipeline

        >>> data, _ = dataretrieval.nwis.get_dv(
        ...     "04288000", parameterCd="00060",
        ...     start="1900-01-01", end="2021-12-31")
        >>> pcts_df = percentiles.calculate_variable_percentile_thresholds_by_day(  # noqa: E501
        ...     data, '00060_Mean',
        ...     percentiles=[0, 5, 10, 25, 75, 90, 95, 100],
        ...     method='linear')
        >>> new_data, _ = dataretrieval.nwis.get_dv(
        ...     "04288000", parameterCd="00060",
        ...     start="2022-05-01", end="2022-05-07")
        >>> new_percentiles = percentiles.calculate_multiple_variable_percentiles_from_values(  # noqa: E501
        ...     new_data, '00060_Mean', pcts_df)
        >>> new_percentiles = utils.categorize_flows(new_percentiles,
        ...     'est_pct', schema_name='NWD')
        >>> new_percentiles[['est_pct', 'flow_cat']].values
        [[13.62, 'Below normal'],
        [14.15, 'Below normal'],
        [14.29, 'Below normal'],
        [23.41, 'Below normal'],
        [27.44, 'Normal'],
        [16.2, 'Below normal'],
        [12.81, 'Below normal']]
    """

    if custom_schema is None:
        schema = retrieve_schema(schema_name)
    else:
        schema = custom_schema

    df['flow_cat'] = pd.cut(df[data_col], schema['ranges'],
                            labels=schema['labels'],
                            include_lowest=True,
                            right=False)
    if "low_label" in schema:
        df['flow_cat'] = df['flow_cat'].cat.add_categories(schema['low_label'])
        df.loc[df[data_col] == schema['ranges'][0], 'flow_cat'] = schema['low_label']  # noqa: E501
        df['flow_cat'] = df['flow_cat'].cat.reorder_categories(
            [schema['low_label']] +
            df['flow_cat'].cat.categories[:-1].tolist())
    if "high_label" in schema:
        df['flow_cat'] = df['flow_cat'].cat.add_categories(schema['high_label'])  # noqa: E501
        df.loc[df[data_col] == schema['ranges'][-1], 'flow_cat'] = schema['high_label']  # noqa: E501

    return df


def retrieve_schema(schema_name):
    """Function used to retrieve the flow range categories given a schema name

    Parameters
    ----------
    schema_name : str
        Name of the categorization schema that should be used to categorize
        streamflow. Available options are 'NWD', 'WaterWatch,
        'WaterWatch_Drought', 'WaterWatch_Flood', 'WaterWatch_BrownBlue', and
        'NIDIS_Drought'.

    Returns
    -------
    schema : dict
        dictionary of flow ranges, category labels, and color palette

    Examples
    --------
    Retrieve the categorization schema 'NWD' to categorization flow similar to
    the USGS National Water Dashboard

    .. doctest::
        :skipif: True

        >>> schema = utils.retrieve_schema('NWD')
        >>> print(schema)
        {'ranges': [0, 10, 25, 76, 90, 100],
        'labels': ['Much below normal', 'Below normal', 'Normal',
            'Above normal', 'Much above normal'],
        'colors': ['#b24249', '#e8ac49', '#44f24e', '#5fd7d9', '#2641f1'],
        'low_label': 'All-time low for this day',
        'low_color': '#e82f3e',
        'high_label': 'All-time high for this day',
        'high_color': '#1f296b'}
    """
    if schema_name.lower() == 'nwd':
        schema = {'ranges': [0, 10, 25, 76, 90, 100],
                  'labels': ['Much below normal', 'Below normal', 'Normal',
                             'Above normal', 'Much above normal'],
                  'colors': ['#b24249', '#e8ac49', '#44f24e', '#5fd7d9',
                             '#2641f1'],
                  'low_label': 'All-time low for this day',
                  'low_color': '#e82f3e',
                  'high_label': 'All-time high for this day',
                  'high_color': "#1f296b"}
    elif schema_name.lower() == 'waterwatch':
        schema = {'ranges': [0, 10, 25, 75, 90, 100],
                  'labels': ['Low', 'Much below normal', 'Below normal',
                             'Normal', 'Above normal',
                             'Much above normal', 'High'],
                  'colors': ['#af2327', '#fda328', '#29fd2f', '#4aded0',
                             '#0b24fb'],
                  'low_label': 'Low',
                  'low_color': '#fc0d1b',
                  'high_label': 'High',
                  'high_color': "#000000"}
    elif schema_name.lower() == 'waterwatch_drought':
        schema = {'ranges': [0, 5, 10, 25],
                  'labels': ['Severe hydrologic drought',
                             'Moderate hydrologic drought',
                             'Below normal'],
                  'colors': ['#af2327', '#fd9941', '#fecb6e'],
                  'low_label': 'Extreme hydrologic drought',
                  'low_color': '#fc0d1b'}
    elif schema_name.lower() == 'waterwatch_flood':
        schema = {'ranges': [0, 95, 99, 101],
                  'labels': ['<95%',
                             '95-98%',
                             '>= 99%'],
                  'colors': ['#ffffff', '#4aded0', '#0b24fb']}
    elif schema_name.lower() == 'waterwatch_brownblue':
        schema = {'ranges': [0, 10, 25, 75, 90, 100],
                  'labels': ['Much below normal', 'Below normal',
                             'Normal', 'Above normal', 'Much above normal'],
                  'colors': ['#dcb668', '#ebd6ab', '#e9e9e9', '#aacee0',
                             '#5699c0'],
                  'low_label': 'Low',
                  'low_color': '#8f4f1f',
                  'high_label': 'High',
                  'high_color': "#292f6b"}
    elif schema_name.lower() == 'nidis_drought':
        schema = {'ranges': [0, 2, 5, 10, 20, 30],
                  'labels': ['Exceptional drought',
                             'Extreme drought',
                             'Severe drought',
                             'Moderate drought',
                             'Abnormally dry'],
                  'colors': ['#720206', '#e30b17', '#fda929', '#fbd285',
                             '#fffd38']}
    else:
        raise ValueError('no matching schema found for ' + schema_name)

    return schema



"""Unit tests for the exceedance module."""
import pytest
import numpy as np
import pandas as pd
from hyswap import exceedance


class TestCalculateExceedanceProbabilities:
    def test_calculate_exceedance_probability_from_distribution(self):
        """Test calculate_exceedance_probability_from_distribution."""
        # lognormal distribution test
        x = 1
        mu = 1
        sigma = 0.25
        prob = exceedance.calculate_exceedance_probability_from_distribution(
            x, 'lognormal', mu, sigma)
        assert prob == 0.6132049428659028

    def test_calculate_exceedance_probability_from_distribution_normal(self):
        # normal distribution test
        x = 1
        mu = 1
        sigma = 0.25
        prob = exceedance.calculate_exceedance_probability_from_distribution(
            x, 'normal', mu, sigma)
        assert prob == 0.5

    def test_calculate_exceedance_probability_from_distribution_weibull(self):
        x = 1
        # Weibull distribution test
        k = 1
        lambda_ = 1
        prob = exceedance.calculate_exceedance_probability_from_distribution(
            x, 'weibull', k, lambda_)
        assert prob == 0.36787944117144233

    def test_calculate_exceedance_probability_from_distribution_exp(self):
        x = 1
        lambda_ = 1
        prob = exceedance.calculate_exceedance_probability_from_distribution(
            x, 'exponential', lambda_)
        assert prob == 1.0

    def test_invalid_exceedance_01(self):
        # Test invalid values
        x = 1
        mu = 1
        sigma = 0.25
        k = 1
        lambda_ = 1
        with pytest.raises(TypeError):
            exceedance.calculate_exceedance_probability_from_distribution(
                x, 'exponential', mu, sigma, k, lambda_)

    def test_invalid_exceedance_02(self):
        # Test invalid values
        x = 1
        mu = 1
        sigma = 0.25
        with pytest.raises(ValueError):
            exceedance.calculate_exceedance_probability_from_distribution(
                x, "invalid", mu, sigma)

    def test_invalid_exceedance_03(self):
        # Test invalid values
        x = 1
        mu = 1
        sigma = 0.25
        with pytest.raises(TypeError):
            exceedance.calculate_exceedance_probability_from_distribution(
                x, 'exponential', mu, sigma, k=1, lambda_=1)

    def test_invalid_exceedance_04(self):
        # Test invalid values
        mu = 1
        sigma = 0.25
        with pytest.raises(TypeError):
            exceedance.calculate_exceedance_probability_from_distribution(
                "invalid", 'exponential', mu, sigma)

    def test_invalid_exceedance_05(self):
        # Test invalid values
        x = 1
        mu = 1
        sigma = 0.25
        with pytest.raises(TypeError):
            exceedance.calculate_exceedance_probability_from_distribution(
                x, 5, mu, sigma)


class TestExceedanceFromValues:
    def test_calculate_exceedance_probability_from_values(self):
        """Test the calculate_exceedance_probability_from_values function."""
        # Testing low value of 1
        x = 1
        values_to_compare = np.array([1, 2, 3, 4])
        prob = exceedance.calculate_exceedance_probability_from_values(
            x, values_to_compare, method='linear')
        assert prob == 1.0

    def test_calculate_exceedance_probability_from_values_list(self):
        values_to_compare = [1, 1, 1, 1]  # use a list
        x = 1
        prob = exceedance.calculate_exceedance_probability_from_values(
            x, values_to_compare, method='linear')
        assert prob == 1.0

    def test_calculate_exceedance_probability_from_values_pandas(self):
        values_to_compare = pd.Series([2, 2, 2, 2])  # use a pandas series
        x = 1
        prob = exceedance.calculate_exceedance_probability_from_values(
            x, values_to_compare, method='linear')
        assert prob == 1.0

    def test_calculate_exceedance_probability_from_values_five(self):
        # Testing high value of 5
        x = 5
        values_to_compare = np.array([1, 2, 3, 4])
        prob = exceedance.calculate_exceedance_probability_from_values(
            x, values_to_compare)
        assert prob == 0.0

    def test_calculate_exceedance_probability_from_values_arr(self):
        values_to_compare = np.array([1, 1, 1, 1])
        x = 5
        prob = exceedance.calculate_exceedance_probability_from_values(
            x, values_to_compare)
        assert prob == 0.0

    def test_calculate_exceedance_probability_from_values_mid(self):
        # Testing middle value of 3
        x = 3
        values_to_compare = np.array([1, 2, 3, 4])
        prob = exceedance.calculate_exceedance_probability_from_values(
            x, values_to_compare)
        assert prob == 0.4

    def test_calculate_exceedance_probability_from_values_mid_arr(self):
        values_to_compare = np.array([1, 1, 1, 1])
        x = 3
        prob = exceedance.calculate_exceedance_probability_from_values(
            x, values_to_compare)
        assert prob == 0.0

    def test_invalid_value_01(self):
        x = 3
        # Testing invalid values
        with pytest.raises(TypeError):
            exceedance.calculate_exceedance_probability_from_values(
                x, "invalid")

    def test_invalid_value_02(self):
        values_to_compare = np.array([1, 1, 1, 1])
        with pytest.raises(TypeError):
            exceedance.calculate_exceedance_probability_from_values(
                "invalid", values_to_compare)

    def test_calculate_exceedance_probability_from_values_weibull(self):
        # Testing weibull method
        x1 = 0.0
        x2 = 2.5
        x3 = 5.0
        values_to_compare = np.array([1, 2, 3, 4])
        prob1 = exceedance.calculate_exceedance_probability_from_values(
            x1, values_to_compare, method='weibull')
        prob2 = exceedance.calculate_exceedance_probability_from_values(
            x2, values_to_compare, method='weibull')
        prob3 = exceedance.calculate_exceedance_probability_from_values(
            x3, values_to_compare, method='weibull')
        assert prob1 == 0.8
        assert prob2 == 0.4
        assert prob3 == 0.0

    def test_calculate_exceedance_probability_from_values_linear(self):
        # Testing linear method
        x1 = 0.0
        x2 = 2.5
        x3 = 5.0
        values_to_compare = np.array([1, 2, 3, 4])
        prob1 = exceedance.calculate_exceedance_probability_from_values(
            x1, values_to_compare, method='linear')
        prob2 = exceedance.calculate_exceedance_probability_from_values(
            x2, values_to_compare, method='linear')
        prob3 = exceedance.calculate_exceedance_probability_from_values(
            x3, values_to_compare, method='linear')
        assert prob1 == 1.0
        assert prob2 == 0.5
        assert prob3 == 0.0

    def test_calculate_exceedance_probability_from_values_hazen(self):
        # Testing hazen method
        x1 = 0.0
        x2 = 2.5
        x3 = 5.0
        values_to_compare = np.array([1, 2, 3, 4])
        prob1 = exceedance.calculate_exceedance_probability_from_values(
            x1, values_to_compare, method='hazen')
        prob2 = exceedance.calculate_exceedance_probability_from_values(
            x2, values_to_compare, method='hazen')
        prob3 = exceedance.calculate_exceedance_probability_from_values(
            x3, values_to_compare, method='hazen')
        assert prob1 == 0.875
        assert prob2 == 0.375
        assert prob3 == -0.125

    def test_calculate_exceedance_probability_from_values_gumbel(self):
        # Testing gumbel method
        x1 = 0.0
        x2 = 2.5
        x3 = 5.0
        values_to_compare = np.array([1, 2, 3, 4])
        prob1 = exceedance.calculate_exceedance_probability_from_values(
            x1, values_to_compare, method='gumbel')
        prob2 = exceedance.calculate_exceedance_probability_from_values(
            x2, values_to_compare, method='gumbel')
        prob3 = exceedance.calculate_exceedance_probability_from_values(
            x3, values_to_compare, method='gumbel')
        assert prob1 == 1.0
        assert prob2 == pytest.approx(1/3)
        assert prob3 == pytest.approx(-1/3)

    def test_calculate_exceedance_probability_from_values_reiss(self):
        # Testing reiss method
        x1 = 0.0
        x2 = 2.5
        x3 = 5.0
        values_to_compare = np.array([1, 2, 3, 4])
        prob1 = exceedance.calculate_exceedance_probability_from_values(
            x1, values_to_compare, method='reiss')
        prob2 = exceedance.calculate_exceedance_probability_from_values(
            x2, values_to_compare, method='reiss')
        prob3 = exceedance.calculate_exceedance_probability_from_values(
            x3, values_to_compare, method='reiss')
        assert prob1 == pytest.approx(0.85, rel=0.01)
        assert prob2 == pytest.approx(0.384, rel=0.01)
        assert prob3 == pytest.approx(-0.077, rel=0.01)

    def test_calculate_exceedance_probability_from_values_blom(self):
        # Testing blom method
        x1 = 0.0
        x2 = 2.5
        x3 = 5.0
        values_to_compare = np.array([1, 2, 3, 4])
        prob1 = exceedance.calculate_exceedance_probability_from_values(
            x1, values_to_compare, method='blom')
        prob2 = exceedance.calculate_exceedance_probability_from_values(
            x2, values_to_compare, method='blom')
        prob3 = exceedance.calculate_exceedance_probability_from_values(
            x3, values_to_compare, method='blom')
        assert prob1 == pytest.approx(0.85, rel=0.01)
        assert prob2 == pytest.approx(0.38, rel=0.01)
        assert prob3 == pytest.approx(-0.088, rel=0.01)

    def test_calculate_exceedance_probability_from_values_error(self):
        # Testing invalid method
        x1 = 0.0
        values_to_compare = np.array([1, 2, 3, 4])
        with pytest.raises(ValueError):
            exceedance.calculate_exceedance_probability_from_values(
                x1, values_to_compare, method='invalid')


class TestExceedanceFromMultiple:
    def test_calculate_exceedance_probability_from_distribution_multiple(self):
        """Test calculate_exceedance_probability_from_distribution_multiple."""
        # Single values should return the same as the single value function
        x = 1
        mu = 1
        sigma = 0.25
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            [x], 'lognormal', mu, sigma)
        prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
            x, 'lognormal', mu, sigma)
        assert prob == prob_single

    def test_calculate_exceedance_probability_from_distribution_multiple_normal(self):  # noqa: E501
        x = 1
        mu = 1
        sigma = 0.25
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            [x], 'normal', mu, sigma)
        prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
            x, 'normal', mu, sigma)
        assert prob == prob_single

    def test_calculate_exceedance_probability_from_distribution_multiple_weibull(self):  # noqa: E501
        x = 1
        k = 1
        lambda_ = 1
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            [x], 'weibull', k, lambda_)
        prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
            x, 'weibull', k, lambda_)
        assert prob == prob_single

    def test_calculate_exceedance_probability_from_distribution_multiple_exp(self):  # noqa: E501
        x = 1
        lambda_ = 1
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            [x], 'exponential', lambda_)
        prob_single = exceedance.calculate_exceedance_probability_from_distribution(  # noqa: E501
            x, 'exponential', lambda_)
        assert prob == prob_single

    def test_calculate_exceedance_probability_from_distribution_multiple_lognorm(self):  # noqa: E501
        # Test multiple values for normal and lognormal distributions
        values = np.array([1.0, 1.25, 1.5, 1.75, 2.0])
        mu = 1
        sigma = 0.25
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            values, 'lognormal', mu, sigma)
        assert np.allclose(prob, np.array([0.61320494, 0.5, 0.41171189,
                                           0.34256783, 0.28787077]))

    def test_calculate_exceedance_probability_from_distribution_multiple_norm01(self):  # noqa: E501
        values = np.array([1.0, 1.25, 1.5, 1.75, 2.0])
        mu = 1
        sigma = 0.25
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            values, 'normal', mu, sigma)
        assert np.allclose(prob, np.array([0.5, 0.158655254, 0.0227501319,
                                           0.00134989803, 0.0000316712418]))

    def test_calculate_exceedance_probability_from_distribution_multiple_weibull01(self):  # noqa: E501
        # Test weibull and exponential distributions
        values = np.array([1, 2, 3, 4])
        k = 1
        lambda_ = 1
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            values, 'weibull', k, lambda_)
        assert np.allclose(prob, np.array([0.36787944, 0.13533528,
                                           0.04978707, 0.01831564]))

    def test_calculate_exceedance_probability_from_distribution_multiple_exp01(self):  # noqa: E501
        values = np.array([1, 2, 3, 4])
        lambda_ = 1
        prob = exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
            values, 'exponential', lambda_)
        assert np.allclose(prob, np.array([1., 0.36787944,
                                           0.13533528, 0.04978707]))

    def test_calculate_exceedance_probability_from_distribution_multiple_type_error(self):  # noqa: E501
        values = np.array([1, 2, 3, 4])
        mu = 1
        sigma = 0.25
        k = 1
        lambda_ = 1
        # Test invalid values
        with pytest.raises(TypeError):
            exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
                values, 'exponential', mu, sigma, k, lambda_)

    def test_calculate_exceedance_probability_from_distribution_multiple_value_error(self):  # noqa: E501
        values = np.array([1, 2, 3, 4])
        mu = 1
        sigma = 0.25
        with pytest.raises(ValueError):
            exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
                values, "invalid", mu, sigma)


class TestExceedanceFromValuesMultiple:
    def test_calculate_exceedance_probability_from_values_multiple_01(self):
        """Test calculate_exceedance_probability_from_values_multiple."""
        values = np.array([1, 2, 3, 4])
        values_to_compare = np.array([1, 2, 3, 4])
        prob = exceedance.calculate_exceedance_probability_from_values_multiple(  # noqa: E501
            values, values_to_compare, method='linear')
        assert np.allclose(prob, np.array([1.0, 0.75, 0.5, 0.25]))

    def test_calculate_exceedance_probability_from_values_multiple_02(self):
        values = np.array([1, 2, 3, 4])
        values_to_compare = np.array([1, 1, 1, 1])
        prob = exceedance.calculate_exceedance_probability_from_values_multiple(  # noqa: E501
            values, values_to_compare, method='linear')
        assert np.allclose(prob, np.array([1.0, 0.0, 0.0, 0.0]))

    def test_calculate_exceedance_probability_from_values_multiple_03(self):
        values = np.array([1, 2, 3, 4])
        values_to_compare = np.array([2, 2, 2, 2])
        prob = exceedance.calculate_exceedance_probability_from_values_multiple(  # noqa: E501
            values, values_to_compare, method='linear')
        assert np.allclose(prob, np.array([1.0, 1.0, 0.0, 0.0]))

    def test_calculate_exceedance_probability_from_values_multiple_04(self):
        values = np.array([1, 2, 3, 4])
        values_to_compare = np.array([1, 2, 3, 4, 1, 2, 3, 4])
        prob = exceedance.calculate_exceedance_probability_from_values_multiple(  # noqa: E501
            values, values_to_compare, method='linear')
        assert np.allclose(prob, np.array([1.0, 0.75, 0.5, 0.25]))

    def test_calculate_exceedance_probability_from_values_multiple_05(self):
        values = np.array([1, 2, 3, 4])
        values_to_compare = np.array([1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4])
        prob = exceedance.calculate_exceedance_probability_from_values_multiple(  # noqa: E501
            values, values_to_compare, method='linear')
        assert np.allclose(prob, np.array([1.0, 0.75, 0.5, 0.25]))



"""Exceedance probability calculations."""

import numpy as np
import pandas as pd
from scipy import stats


def calculate_exceedance_probability_from_distribution(x, dist,
                                                       *args, **kwargs):
    """
    Calculate the exceedance probability of a value relative to a distribution.

    Parameters
    ----------
    x : float
        The value for which to calculate the exceedance probability.
    dist : str
        The distribution to use. Must be one of 'lognormal', 'normal',
        'weibull', or 'exponential'.
    *args
        Positional arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.
    **kwargs
        Keyword arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.

    Returns
    -------
    float
        The exceedance probability.

    Examples
    --------
    Calculating the exceedance probability of a value of 1 from a lognormal
    distribution with a mean of 1 and a standard deviation of 0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution(
        ...     1, 'lognormal', 1, 0.25)
        0.6132049428659028

    Calculating the exceedance probability of a value of 1 from a normal
    distribution with a mean of 1 and a standard deviation of 0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution(
        ...     1, 'normal', 1, 0.25)
        0.5
    """
    # type check
    if not isinstance(x, (int, float, np.int64, np.float64, np.float32,
                          np.float16, np.int32, np.int16, np.int8)):
        raise TypeError("x must be a float (or integer).")
    if not isinstance(dist, str):
        raise TypeError("dist must be a string.")
    # do calculation
    if dist == 'lognormal':
        return stats.lognorm.sf(x, *args, **kwargs)
    elif dist == 'normal':
        return stats.norm.sf(x, *args, **kwargs)
    elif dist == 'weibull':
        return stats.exponweib.sf(x, *args, **kwargs)
    elif dist == 'exponential':
        return stats.expon.sf(x, *args, **kwargs)
    else:
        raise ValueError("dist must be one of 'lognormal', 'normal'," +
                         "'weibull', or 'exponential'.")


def calculate_exceedance_probability_from_values(x, values_to_compare,
                                                 method="weibull"):
    """
    Calculate the exceedance probability of a value compared to several values.

    This function computes an exceedance probability using common plotting
    position formulas, with the default being the 'Weibull' method (also known
    as Type 6 in R). The value (x) is ranked among the values to compare by
    determining the number that are *greater than or equal to* the
    input value (x), which provides the minimum rank in the case of tied
    values. Additional methods other than the 'Weibull' method can be specified
    and are described in more detail in Helsel et al 2020.

    Helsel, D.R., Hirsch, R.M., Ryberg, K.R., Archfield, S.A.,
      and Gilroy, E.J., 2020, Statistical methods in water resources:
      U.S. Geological Survey Techniques and Methods, book 4, chap. A3,
      458 p., https://doi.org/10.3133/tm4a3. [Supersedes USGS Techniques
      of Water-Resources Investigations, book 4, chap. A3, version 1.1.]

    Parameters
    ----------
    x : float
        The value for which to calculate the exceedance probability.
    values_to_compare : array-like
        The values to use to calculate the exceedance probability.

    method : str, optional
        Method (formulation) of plotting position formula.
        Default is 'weibull' (Type 6).

    Returns
    -------
    float
        The exceedance probability.

    Examples
    --------
    Calculating the exceedance probability of a value of 1 from a set of values
    of 1, 2, 3, and 4.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_values(
        ...     1, [1, 2, 3, 4], method='linear')
        1.0

    Calculating the exceedance probability of a value of 5 from a set of values
    of 1, 2, 3, and 4.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_values(
        ...     5, [1, 2, 3, 4])
        0.0
    """

    if method == 'weibull':     # R Type 6
        alpha = 0
        beta = 0
    elif method == 'linear':    # R Type 4
        alpha = 0
        beta = 1
    elif method == 'hazen':     # R Type 5
        alpha = 0.5
        beta = 0.5
    elif method == 'gumbel':    # R Type 7
        alpha = 1
        beta = 1
    elif method == "reiss":     # R Type 8
        alpha = 1/3
        beta = 1/3
    elif method == "blom":      # R Type 9
        alpha = 3/8
        beta = 3/8
    else:
        raise ValueError("dist must be one of 'weibull', 'linear'," +
                         "'hazen', 'gumbel', 'reiss', or 'blom'.")

    # some type conversions to get to a numpy array
    if isinstance(values_to_compare, list):
        values_to_compare = np.array(values_to_compare)
    elif isinstance(values_to_compare, pd.Series):
        values_to_compare = values_to_compare.values
    # raise error if not a numpy array
    if not isinstance(values_to_compare, np.ndarray):
        raise TypeError("values_to_compare must be a numpy array, list, " +
                        "or pandas Series.")
    # calculate the exceedance probability
    exceed_prob = (np.sum(values_to_compare >= x) - alpha) / (len(values_to_compare) + 1 - alpha - beta)  # noqa: E501

    return exceed_prob


def calculate_exceedance_probability_from_distribution_multiple(values, dist,
                                                                *args,
                                                                **kwargs):
    """
    Calculate the exceedance probability of multiple values vs a distribution.

    Parameters
    ----------
    values : array-like
        The values for which to calculate the exceedance probability.
    dist : str
        The distribution to use. Must be one of 'lognormal', 'normal',
        'weibull', or 'exponential'.
    *args
        Positional arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.
    **kwargs
        Keyword arguments to pass to the distribution, which is one of
        `stats.lognorm.sf`, `stats.norm.sf`, `stats.exponweib.sf` or
        `stats.expon.sf`, refer to the `scipy.stats` documentation for more
        information about these arguments.

    Returns
    -------
    array-like
        The exceedance probabilities.

    Examples
    --------
    Calculating the exceedance probability of a set of values of 1, 1.25 and
    1.5 from a lognormal distribution with a mean of 1 and a standard
    deviation of 0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        ...     [1, 1.25, 1.5], 'lognormal', 1, 0.25)
        array([0.61320494, 0.5       , 0.41171189])

    Calculating the exceedance probability of a set of values of 1, 2, 3, and 4
    from a normal distribution with a mean of 1 and a standard deviation of
    0.25.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_distribution_multiple(  # noqa: E501
        ...     [1, 2, 3, 4], 'normal', 1, 0.25)
        array([5.00000000e-01, 3.16712418e-05, 6.22096057e-16, 1.77648211e-33])
    """
    return np.array([calculate_exceedance_probability_from_distribution(
        x, dist, *args, **kwargs) for x in values])


def calculate_exceedance_probability_from_values_multiple(values,
                                                          values_to_compare,
                                                          method="weibull"):
    """
    Calculate the exceedance probability of multiple values vs a set of values.
    All methods supported in *calculate_exceedance_probability_from_values*
    are supported and by default uses the 'Weibull' method.

    Parameters
    ----------
    values : array-like
        The values for which to calculate the exceedance probability.
    values_to_compare : array-like
        The values to use to calculate the exceedance probability.
    method : str, optional
        Method (formulation) of plotting position formula.
        Default is 'weibull' (Type 6).

    Returns
    -------
    array-like
        The exceedance probabilities.

    Examples
    --------
    Calculating the exceedance probability of a set of values of 1, 1.25 and
    1.5 from a set of values of 1, 2, 3, and 4.

    .. doctest::

        >>> exceedance.calculate_exceedance_probability_from_values_multiple(
        ...     [1, 1.25, 2.5], [1, 2, 3, 4], method='linear')
        array([1.  , 0.75, 0.5 ])

    Fetch some data from NWIS and calculate the exceedance probability for a
    set of 5 values spaced evenly between the minimum and maximum values.

    .. doctest::
        :skipif: True  # skips this block of code as it broke CI pipeline

        >>> df, _ = dataretrieval.nwis.get_gwlevels(site='434400121275801',
        ...                                         start='2000-01-01',
        ...                                         end='2020-01-01')
        >>> values = np.linspace(df['lev_va'].min(),
        ...                      df['lev_va'].max(), 5)
        >>> exceedance.calculate_exceedance_probability_from_values_multiple(
        ...     values, df['lev_va'])
        array([1.        , 0.96363636, 0.83636364, 0.47272727, 0.01818182])
    """
    return np.array([calculate_exceedance_probability_from_values(
        x, values_to_compare, method=method) for x in values])
